"""Unit test for Quantum Composers 9530 Pulse Generator."""

import logging
from typing import cast
import unittest
from unittest.mock import MagicMock, call, patch

import qmi
from qmi.core.exceptions import QMI_InstrumentException
from qmi.core.transport import QMI_TcpTransport
from qmi.instruments.quantum_composers import QuantumComposers_9530
from qmi.instruments.quantum_composers import (
    RefClkSource, PulseMode, TriggerMode, TriggerEdge, OutputDriver
)


class SuppressLogging:
    """Context manager to temporarily suppress logging during a test."""

    def __enter__(self):
        # Suppress logging of all levels up to ERROR.
        logging.getLogger("qmi.instruments.quantum_composers.pulse_generator9530").setLevel(logging.CRITICAL)

    def __exit__(self, typ, value, tb):
        # Restore default log levels.
        logging.getLogger("qmi.instruments.quantum_composers.pulse_generator9530").setLevel(logging.NOTSET)


class TestPulseGenerator9530(unittest.TestCase):

    def setUp(self):
        qmi.start("TestContext")
        self._transport_mock = MagicMock(spec=QMI_TcpTransport)
        with patch(
                'qmi.instruments.quantum_composers.pulse_generator9530.create_transport',
                return_value=self._transport_mock):
            self.instr: QuantumComposers_9530 = qmi.make_instrument("instr", QuantumComposers_9530, "transport_descriptor")
            self.instr = cast(QuantumComposers_9530, self.instr)

    def tearDown(self):
        qmi.stop()

    def test_open_close(self):
        self.instr.open()
        self._transport_mock.open.assert_called_once_with()
        self.instr.close()
        self._transport_mock.close.assert_called_once_with()

    def test_get_idn(self):
        self._transport_mock.read_until.return_value = b"vendor,model,serial,version\r\n"
        self.instr.open()
        idn = self.instr.get_idn()
        self.instr.close()
        self._transport_mock.write.assert_called_once_with(b"*IDN?\r\n")
        self.assertEqual(idn.vendor, "vendor")
        self.assertEqual(idn.model, "model")
        self.assertEqual(idn.serial, "serial")
        self.assertEqual(idn.version, "version")

    def test_get_idn_fail(self):
        self._transport_mock.read_until.return_value = b"nonsense\r\n"
        self.instr.open()
        with SuppressLogging(), self.assertRaises(QMI_InstrumentException):
            idn = self.instr.get_idn()
        self.instr.close()
        self._transport_mock.write.assert_called_once_with(b"*IDN?\r\n")

    def test_reset(self):
        self._transport_mock.read_until.return_value = b"ok\r\n"
        self.instr.open()
        self.instr.reset()
        self.instr.close()
        self._transport_mock.write.assert_called_once_with(b"*RST\r\n")

    def test_get_num_channels(self):
        self._transport_mock.read_until.return_value = b"T0, CHA, CHB, CHC, CHD\r\n"
        self.instr.open()
        num_channels = self.instr.get_num_channels()
        self.instr.close()
        self.assertEqual(num_channels, 4)
        self._transport_mock.write.assert_called_once_with(b":INST:CAT?\r\n")

    def test_get_num_channels_fail(self):
        self._transport_mock.read_until.return_value = b"?3\r\n"
        self.instr.open()
        with SuppressLogging(), self.assertRaises(QMI_InstrumentException):
            num_channels = self.instr.get_num_channels()
        self.instr.close()
        self._transport_mock.write.assert_called_once_with(b":INST:CAT?\r\n")

    def test_get_refclk_source(self):
        self.instr.open()

        self._transport_mock.read_until.return_value = b"INT\r\n"
        source = self.instr.get_refclk_source()
        self.assertEqual(source, RefClkSource.INTERNAL)

        self._transport_mock.read_until.return_value = b"EXT\r\n"
        source = self.instr.get_refclk_source()
        self.assertEqual(source, RefClkSource.EXTERNAL)

        self._transport_mock.read_until.return_value = b"XPL\r\n"
        source = self.instr.get_refclk_source()
        self.assertEqual(source, RefClkSource.EXTPLL)

        self._transport_mock.read_until.return_value = b"XPL?\r\n"
        source = self.instr.get_refclk_source()
        self.assertEqual(source, RefClkSource.EXTPLL)

        self.instr.close()
        self._transport_mock.write.assert_has_calls([
            call(b":PULSE0:ICL:MOD?\r\n"),
            call(b":PULSE0:ICL:MOD?\r\n"),
            call(b":PULSE0:ICL:MOD?\r\n"),
            call(b":PULSE0:ICL:MOD?\r\n"),
        ])

    def test_get_refclk_source_fail(self):
        self._transport_mock.read_until.return_value = b"unknown\r\n"
        self.instr.open()
        with self.assertRaises(QMI_InstrumentException):
            source = self.instr.get_refclk_source()
        self.instr.close()

    def test_is_refclk_pll_locked(self):
        self.instr.open()

        self._transport_mock.read_until.return_value = b"XPL\r\n"
        locked = self.instr.is_refclk_pll_locked()
        self.assertTrue(locked)

        self._transport_mock.read_until.return_value = b"XPL?\r\n"
        locked = self.instr.is_refclk_pll_locked()
        self.assertFalse(locked)

        self._transport_mock.read_until.return_value = b"INT\r\n"
        locked = self.instr.is_refclk_pll_locked()
        self.assertFalse(locked)

        self.instr.close()
        self._transport_mock.write.assert_has_calls([
            call(b":PULSE0:ICL:MOD?\r\n"),
            call(b":PULSE0:ICL:MOD?\r\n"),
            call(b":PULSE0:ICL:MOD?\r\n"),
        ])

    def test_set_refclk_source(self):
        self._transport_mock.read_until.return_value = b"ok\r\n"
        self.instr.open()
        self.instr.set_refclk_source(RefClkSource.INTERNAL)
        self.instr.set_refclk_source(RefClkSource.EXTERNAL)
        self.instr.set_refclk_source(RefClkSource.EXTPLL)
        self.instr.close()
        self._transport_mock.write.assert_has_calls([
            call(b":PULSE0:ICL:MOD INT\r\n"),
            call(b":PULSE0:ICL:MOD EXT\r\n"),
            call(b":PULSE0:ICL:MOD XPL\r\n"),
        ])

    def test_get_refclk_rate(self):
        self._transport_mock.read_until.return_value = b"50\r\n"
        self.instr.open()
        rate = self.instr.get_refclk_rate()
        self.instr.close()
        self.assertEqual(rate, 50)
        self._transport_mock.write.assert_called_once_with(b":PULSE0:ICL:RAT?\r\n")

    def test_bad_int_response(self):
        self._transport_mock.read_until.return_value = b"not_int\r\n"
        self.instr.open()
        with SuppressLogging(), self.assertRaises(QMI_InstrumentException):
            rate = self.instr.get_refclk_rate()
        self.instr.close()
        self._transport_mock.write.assert_called_once_with(b":PULSE0:ICL:RAT?\r\n")

    def test_set_refclk_rate(self):
        self._transport_mock.read_until.return_value = b"ok\r\n"
        self.instr.open()
        self.instr.set_refclk_rate(10)
        self.instr.close()
        self._transport_mock.write.assert_called_once_with(b":PULSE0:ICL:RAT 10\r\n")

    def test_get_refclk_level(self):
        self._transport_mock.read_until.return_value = b"0.50\r\n"
        self.instr.open()
        level = self.instr.get_refclk_level()
        self.instr.close()
        self.assertEqual(level, 0.5)
        self._transport_mock.write.assert_called_once_with(b":PULSE0:ICL:LEV?\r\n")

    def test_bad_float_response(self):
        self._transport_mock.read_until.return_value = b"not_float\r\n"
        self.instr.open()
        with SuppressLogging(), self.assertRaises(QMI_InstrumentException):
            level = self.instr.get_refclk_level()
        self.instr.close()
        self._transport_mock.write.assert_called_once_with(b":PULSE0:ICL:LEV?\r\n")

    def test_set_refclk_level(self):
        self._transport_mock.read_until.return_value = b"ok\r\n"
        self.instr.open()
        self.instr.set_refclk_level(1.5)
        self.instr.close()
        self._transport_mock.write.assert_called_once_with(b":PULSE0:ICL:LEV 1.500\r\n")

    def test_set_refclk_external(self):
        self._transport_mock.read_until.return_value = b"ok\r\n"
        self.instr.open()
        self.instr.set_refclk_external(rate=10, level=1.0, use_pll=True)
        self.instr.close()
        self._transport_mock.write.assert_has_calls([
            call(b":PULSE0:ICL:MOD XPL\r\n"),
            call(b":PULSE0:ICL:RAT 10\r\n"),
            call(b":PULSE0:ICL:LEV 1.000\r\n"),
        ])

    def test_get_output_enabled(self):
        self.instr.open()

        self._transport_mock.read_until.return_value = b"1\r\n"
        enabled = self.instr.get_output_enabled()
        self.assertTrue(enabled)

        self._transport_mock.read_until.return_value = b"0\r\n"
        enabled = self.instr.get_output_enabled()
        self.assertFalse(enabled)

        self.instr.close()
        self._transport_mock.write.assert_has_calls([
            call(b":PULSE0:STAT?\r\n"),
            call(b":PULSE0:STAT?\r\n"),
        ])

    def test_bad_bool_response(self):
        self._transport_mock.read_until.return_value = b"maybe\r\n"
        self.instr.open()
        with SuppressLogging(), self.assertRaises(QMI_InstrumentException):
            enabled = self.instr.get_output_enabled()
        self.instr.close()
        self._transport_mock.write.assert_called_once_with(b":PULSE0:STAT?\r\n")

    def test_set_output_enabled(self):
        self._transport_mock.read_until.return_value = b"ok\r\n"
        self.instr.open()
        self.instr.set_output_enabled(True)
        self.instr.set_output_enabled(False)
        self.instr.close()
        self._transport_mock.write.assert_has_calls([
            call(b":PULSE0:STAT 1\r\n"),
            call(b":PULSE0:STAT 0\r\n"),
        ])

    def test_set_output_enabled_fail(self):
        self._transport_mock.read_until.return_value = b"?3\r\n"
        self.instr.open()
        with SuppressLogging(), self.assertRaises(QMI_InstrumentException):
            self.instr.set_output_enabled(True)
        self.instr.close()
        self._transport_mock.write.assert_called_once_with(b":PULSE0:STAT 1\r\n")

    def test_set_output_enabled_fail_nonsense(self):
        self.instr.open()
        self._transport_mock.read_until.return_value = b"?unknown_error\r\n"
        with SuppressLogging(), self.assertRaises(QMI_InstrumentException):
            self.instr.set_output_enabled(True)
        self._transport_mock.read_until.return_value = b"?100\r\n"
        with SuppressLogging(), self.assertRaises(QMI_InstrumentException):
            self.instr.set_output_enabled(True)
        self.instr.close()

    def test_get_t0_period(self):
        self._transport_mock.read_until.return_value = b"2.718281828\r\n"
        self.instr.open()
        period = self.instr.get_t0_period()
        self.instr.close()
        self.assertEqual(period, 2.718281828)
        self._transport_mock.write.assert_called_once_with(b":PULSE0:PER?\r\n")

    def test_set_t0_period(self):
        self._transport_mock.read_until.return_value = b"ok\r\n"
        self.instr.open()
        self.instr.set_t0_period(3.141592653589793)
        self.instr.close()
        self._transport_mock.write.assert_called_once_with(b":PULSE0:PER 3.141592654\r\n")

    def test_set_t0_period(self):
        self._transport_mock.read_until.return_value = b"ok\r\n"
        self.instr.open()
        self.instr.set_t0_period(3.141592653589793)
        self.instr.close()
        self._transport_mock.write.assert_called_once_with(b":PULSE0:PER 3.141592654\r\n")

    def test_get_t0_mode(self):
        self.instr.open()

        self._transport_mock.read_until.return_value = b"NORM\r\n"
        mode = self.instr.get_t0_mode()
        self.assertEqual(mode, PulseMode.NORMAL)

        self._transport_mock.read_until.return_value = b"SING\r\n"
        mode = self.instr.get_t0_mode()
        self.assertEqual(mode, PulseMode.SINGLE)

        self._transport_mock.read_until.return_value = b"BURS\r\n"
        mode = self.instr.get_t0_mode()
        self.assertEqual(mode, PulseMode.BURST)

        self._transport_mock.read_until.return_value = b"DCYC\r\n"
        mode = self.instr.get_t0_mode()
        self.assertEqual(mode, PulseMode.DUTYCYCLE)

        self.instr.close()
        self._transport_mock.write.assert_has_calls([
            call(b":PULSE0:MOD?\r\n"),
            call(b":PULSE0:MOD?\r\n"),
            call(b":PULSE0:MOD?\r\n"),
        ])

    def test_get_t0_mode_fail(self):
        self._transport_mock.read_until.return_value = b"unknown\r\n"
        self.instr.open()
        with self.assertRaises(QMI_InstrumentException):
            mode = self.instr.get_t0_mode()
        self.instr.close()

    def test_set_t0_mode(self):
        self._transport_mock.read_until.return_value = b"ok\r\n"
        self.instr.open()
        self.instr.set_t0_mode(PulseMode.NORMAL)
        self.instr.set_t0_mode(PulseMode.SINGLE)
        self.instr.set_t0_mode(PulseMode.BURST)
        self.instr.set_t0_mode(PulseMode.DUTYCYCLE)
        self.instr.close()
        self._transport_mock.write.assert_has_calls([
            call(b":PULSE0:MOD NORM\r\n"),
            call(b":PULSE0:MOD SING\r\n"),
            call(b":PULSE0:MOD BURS\r\n"),
            call(b":PULSE0:MOD DCYC\r\n"),
        ])

    def test_get_t0_burst_count(self):
        self._transport_mock.read_until.return_value = b"5\r\n"
        self.instr.open()
        burst = self.instr.get_t0_burst_count()
        self.instr.close()
        self.assertEqual(burst, 5)
        self._transport_mock.write.assert_called_once_with(b":PULSE0:BCO?\r\n")

    def test_set_t0_burst_count(self):
        self._transport_mock.read_until.return_value = b"ok\r\n"
        self.instr.open()
        self.instr.set_t0_burst_count(21)
        self.instr.close()
        self._transport_mock.write.assert_called_once_with(b":PULSE0:BCO 21\r\n")

    def test_get_t0_duty_cycle(self):
        self._transport_mock.read_until.side_effect = [
            b"3\r\n",
            b"7\r\n"
        ]
        self.instr.open()
        duty_cycle = self.instr.get_t0_duty_cycle()
        self.instr.close()
        self.assertEqual(duty_cycle, (3, 7))
        self._transport_mock.write.assert_has_calls([
            call(b":PULSE0:PCO?\r\n"),
            call(b":PULSE0:OCO?\r\n"),
        ])

    def test_set_t0_duty_cycle(self):
        self._transport_mock.read_until.return_value = b"ok\r\n"
        self.instr.open()
        self.instr.set_t0_duty_cycle(5, 11)
        self.instr.close()
        self._transport_mock.write.assert_has_calls([
            call(b":PULSE0:PCO 5\r\n"),
            call(b":PULSE0:OCO 11\r\n"),
        ])

    def test_get_trigger_mode(self):
        self.instr.open()

        self._transport_mock.read_until.return_value = b"DIS\r\n"
        mode = self.instr.get_trigger_mode()
        self.assertEqual(mode, TriggerMode.DISABLED)

        self._transport_mock.read_until.return_value = b"TRIG\r\n"
        mode = self.instr.get_trigger_mode()
        self.assertEqual(mode, TriggerMode.ENABLED)

        self._transport_mock.read_until.return_value = b"DUAL\r\n"
        mode = self.instr.get_trigger_mode()
        self.assertEqual(mode, TriggerMode.DUAL)

        self.instr.close()
        self._transport_mock.write.assert_has_calls([
            call(b":PULSE0:TRIG:MOD?\r\n"),
            call(b":PULSE0:TRIG:MOD?\r\n"),
            call(b":PULSE0:TRIG:MOD?\r\n"),
        ])

    def test_get_trigger_mode_fail(self):
        self._transport_mock.read_until.return_value = b"unknown\r\n"
        self.instr.open()
        with self.assertRaises(QMI_InstrumentException):
            mode = self.instr.get_trigger_mode()
        self.instr.close()

    def test_set_trigger_mode(self):
        self._transport_mock.read_until.return_value = b"ok\r\n"
        self.instr.open()
        self.instr.set_trigger_mode(TriggerMode.DISABLED)
        self.instr.set_trigger_mode(TriggerMode.ENABLED)
        self.instr.close()
        self._transport_mock.write.assert_has_calls([
            call(b":PULSE0:TRIG:MOD DIS\r\n"),
            call(b":PULSE0:TRIG:MOD TRIG\r\n"),
        ])

    def test_get_trigger_edge(self):
        self.instr.open()

        self._transport_mock.read_until.return_value = b"RIS\r\n"
        mode = self.instr.get_trigger_edge()
        self.assertEqual(mode, TriggerEdge.RISING)

        self._transport_mock.read_until.return_value = b"FALL\r\n"
        mode = self.instr.get_trigger_edge()
        self.assertEqual(mode, TriggerEdge.FALLING)

        self.instr.close()
        self._transport_mock.write.assert_has_calls([
            call(b":PULSE0:TRIG:EDGE?\r\n"),
            call(b":PULSE0:TRIG:EDGE?\r\n"),
        ])

    def test_get_trigger_edge_fail(self):
        self._transport_mock.read_until.return_value = b"unknown\r\n"
        self.instr.open()
        with self.assertRaises(QMI_InstrumentException):
            edge = self.instr.get_trigger_edge()
        self.instr.close()

    def test_set_trigger_edge(self):
        self._transport_mock.read_until.return_value = b"ok\r\n"
        self.instr.open()
        self.instr.set_trigger_edge(TriggerEdge.RISING)
        self.instr.set_trigger_edge(TriggerEdge.FALLING)
        self.instr.close()
        self._transport_mock.write.assert_has_calls([
            call(b":PULSE0:TRIG:EDGE RIS\r\n"),
            call(b":PULSE0:TRIG:EDGE FALL\r\n"),
        ])

    def test_get_trigger_level(self):
        self._transport_mock.read_until.return_value = b"0.90\r\n"
        self.instr.open()
        level = self.instr.get_trigger_level()
        self.instr.close()
        self.assertEqual(level, 0.90)
        self._transport_mock.write.assert_called_once_with(b":PULSE0:TRIG:LEVEL?\r\n")

    def test_set_trigger_level(self):
        self._transport_mock.read_until.return_value = b"ok\r\n"
        self.instr.open()
        self.instr.set_trigger_level(1.23)
        self.instr.close()
        self._transport_mock.write.assert_called_once_with(b":PULSE0:TRIG:LEVEL 1.230\r\n")

    def test_get_channel_enabled(self):
        self._transport_mock.read_until.return_value = b"0\r\n"
        self.instr.open()
        enabled = self.instr.get_channel_enabled(3)
        self.instr.close()
        self.assertFalse(enabled)
        self._transport_mock.write.assert_called_once_with(b":PULSE3:STAT?\r\n")

    def test_set_channel_enabled(self):
        self._transport_mock.read_until.return_value = b"ok\r\n"
        self.instr.open()
        self.instr.set_channel_enabled(4, True)
        self.instr.close()
        self._transport_mock.write.assert_called_once_with(b":PULSE4:STAT 1\r\n")

    def test_get_channel_width(self):
        self._transport_mock.read_until.return_value = b"0.01234567891\r\n"
        self.instr.open()
        width = self.instr.get_channel_width(5)
        self.instr.close()
        self.assertEqual(width, 0.01234567891)
        self._transport_mock.write.assert_called_once_with(b":PULSE5:WIDT?\r\n")

    def test_set_channel_width(self):
        self._transport_mock.read_until.return_value = b"ok\r\n"
        self.instr.open()
        self.instr.set_channel_width(6, 1.23456789125)
        self.instr.close()
        self._transport_mock.write.assert_called_once_with(b":PULSE6:WIDT 1.23456789125\r\n")

    def test_get_channel_delay(self):
        self._transport_mock.read_until.return_value = b"0.23456789123\r\n"
        self.instr.open()
        width = self.instr.get_channel_delay(7)
        self.instr.close()
        self.assertEqual(width, 0.23456789123)
        self._transport_mock.write.assert_called_once_with(b":PULSE7:DEL?\r\n")

    def test_set_channel_delay(self):
        self._transport_mock.read_until.return_value = b"ok\r\n"
        self.instr.open()
        self.instr.set_channel_delay(8, 0.34567891275)
        self.instr.close()
        self._transport_mock.write.assert_called_once_with(b":PULSE8:DEL 0.34567891275\r\n")

    def test_get_channel_mode(self):
        self._transport_mock.read_until.return_value = b"NORM\r\n"
        self.instr.open()
        mode = self.instr.get_channel_mode(1)
        self.instr.close()
        self.assertEqual(mode, PulseMode.NORMAL)
        self._transport_mock.write.assert_called_once_with(b":PULSE1:CMOD?\r\n")

    def test_get_channel_mode_fail(self):
        self._transport_mock.read_until.return_value = b"unknown\r\n"
        self.instr.open()
        with self.assertRaises(QMI_InstrumentException):
            edge = self.instr.get_channel_mode(1)
        self.instr.close()

    def test_set_channel_mode(self):
        self._transport_mock.read_until.return_value = b"ok\r\n"
        self.instr.open()
        self.instr.set_channel_mode(2, PulseMode.BURST)
        self.instr.close()
        self._transport_mock.write.assert_called_once_with(b":PULSE2:CMOD BURS\r\n")

    def test_get_channel_burst_count(self):
        self._transport_mock.read_until.return_value = b"11\r\n"
        self.instr.open()
        burst = self.instr.get_channel_burst_count(3)
        self.instr.close()
        self.assertEqual(burst, 11)
        self._transport_mock.write.assert_called_once_with(b":PULSE3:BCO?\r\n")

    def test_set_channel_burst_count(self):
        self._transport_mock.read_until.return_value = b"ok\r\n"
        self.instr.open()
        self.instr.set_channel_burst_count(3, 4)
        self.instr.close()
        self._transport_mock.write.assert_called_once_with(b":PULSE3:BCO 4\r\n")

    def test_get_channel_duty_cycle(self):
        self._transport_mock.read_until.side_effect = [
            b"8\r\n",
            b"9\r\n",
        ]
        self.instr.open()
        duty_cycle = self.instr.get_channel_duty_cycle(1)
        self.instr.close()
        self.assertEqual(duty_cycle, (8, 9))
        self._transport_mock.write.assert_has_calls([
            call(b":PULSE1:PCO?\r\n"),
            call(b":PULSE1:OCO?\r\n"),
        ])

    def test_set_channel_duty_cycle(self):
        self._transport_mock.read_until.return_value = b"ok\r\n"
        self.instr.open()
        self.instr.set_channel_duty_cycle(1, 10, 11)
        self.instr.close()
        self._transport_mock.write.assert_has_calls([
            call(b":PULSE1:PCO 10\r\n"),
            call(b":PULSE1:OCO 11\r\n"),
        ])

    def test_get_output_driver(self):
        self.instr.open()

        self._transport_mock.read_until.return_value = b"TTL\r\n"
        driver = self.instr.get_output_driver(2)
        self.assertEqual(driver, OutputDriver.TTL)

        self._transport_mock.read_until.return_value = b"ADJ\r\n"
        driver = self.instr.get_output_driver(3)
        self.assertEqual(driver, OutputDriver.ADJUSTABLE)

        self.instr.close()
        self._transport_mock.write.assert_has_calls([
            call(b":PULSE2:OUTP:MODE?\r\n"),
            call(b":PULSE3:OUTP:MODE?\r\n"),
        ])

    def test_set_output_driver(self):
        self._transport_mock.read_until.return_value = b"ok\r\n"
        self.instr.open()
        self.instr.set_output_driver(2, OutputDriver.TTL)
        self.instr.set_output_driver(3, OutputDriver.ADJUSTABLE)
        self.instr.close()
        self._transport_mock.write.assert_has_calls([
            call(b":PULSE2:OUTP:MODE TTL\r\n"),
            call(b":PULSE3:OUTP:MODE ADJ\r\n"),
        ])

    def test_get_output_ampliude(self):
        self._transport_mock.read_until.return_value = b"3.14\r\n"
        self.instr.open()
        ampl = self.instr.get_output_amplitude(1)
        self.instr.close()
        self.assertEqual(ampl, 3.14)
        self._transport_mock.write.assert_called_once_with(b":PULSE1:OUTP:AMPL?\r\n")

    def test_set_output_ampliude(self):
        self._transport_mock.read_until.return_value = b"ok\r\n"
        self.instr.open()
        self.instr.set_output_amplitude(2, 3.3)
        self.instr.close()
        self._transport_mock.write.assert_called_once_with(b":PULSE2:OUTP:AMPL 3.300\r\n")

    def test_get_output_inverted(self):
        self._transport_mock.read_until.return_value = b"INV\r\n"
        self.instr.open()
        inv = self.instr.get_output_inverted(1)
        self.instr.close()
        self.assertTrue(inv)
        self._transport_mock.write.assert_called_once_with(b":PULSE1:POL?\r\n")

    def test_set_output_inverted(self):
        self._transport_mock.read_until.return_value = b"ok\r\n"
        self.instr.open()
        self.instr.set_output_inverted(2, False)
        self.instr.close()
        self._transport_mock.write.assert_called_once_with(b":PULSE2:POL NORM\r\n")

    def test_bad_channel(self):
        self.instr.open()
        with self.assertRaises(ValueError):
            self.instr.get_channel_enabled(0)
        with self.assertRaises(ValueError):
            self.instr.set_channel_enabled(0, True)
        with self.assertRaises(ValueError):
            self.instr.get_channel_width(0)
        with self.assertRaises(ValueError):
            self.instr.set_channel_width(0, 0.1)
        with self.assertRaises(ValueError):
            self.instr.get_channel_delay(0)
        with self.assertRaises(ValueError):
            self.instr.set_channel_delay(0, 0.1)
        with self.assertRaises(ValueError):
            self.instr.get_channel_mode(0)
        with self.assertRaises(ValueError):
            self.instr.set_channel_mode(0, PulseMode.NORMAL)
        with self.assertRaises(ValueError):
            self.instr.get_channel_burst_count(0)
        with self.assertRaises(ValueError):
            self.instr.set_channel_burst_count(0, 1)
        with self.assertRaises(ValueError):
            self.instr.get_channel_duty_cycle(0)
        with self.assertRaises(ValueError):
            self.instr.set_channel_duty_cycle(0, 1, 2)
        with self.assertRaises(ValueError):
            self.instr.get_output_driver(0)
        with self.assertRaises(ValueError):
            self.instr.set_output_driver(0, OutputDriver.TTL)
        with self.assertRaises(ValueError):
            self.instr.get_output_amplitude(0)
        with self.assertRaises(ValueError):
            self.instr.set_output_amplitude(0, 3.0)
        with self.assertRaises(ValueError):
            self.instr.get_output_inverted(0)
        with self.assertRaises(ValueError):
            self.instr.set_output_inverted(0, False)
        self.instr.close()

    def test_out_of_range(self):
        self.instr.open()
        with self.assertRaises(ValueError):
            self.instr.set_t0_period(0)
        with self.assertRaises(ValueError):
            self.instr.set_t0_period(10000)
        self.instr.close()

    def test_set_display_enabled(self):
        self._transport_mock.read_until.return_value = b"ok\r\n"
        self.instr.open()
        self.instr.set_display_enabled(False)
        self.instr.close()
        self._transport_mock.write.assert_called_once_with(b":DISP:ENAB 0\r\n")


if __name__ == '__main__':
    unittest.main()



"""Instrument driver for the OzOptics Electric Polarization Controller driver."""
import re
from typing import Tuple, Union

from qmi.core.context import QMI_Context
from qmi.core.exceptions import QMI_InstrumentException
from qmi.core.instrument import QMI_Instrument
from qmi.core.rpc import rpc_method
from qmi.core.transport import QMI_Transport, create_transport


class OzOptics_EpcDriver(QMI_Instrument):
    """Instrument driver for the OzOptics Electric Polarization Controller driver."""

    def __init__(self, context: QMI_Context, name: str, transport: Union[str, QMI_Transport]) -> None:
        """Initialize the Electric Polarization Controller driver.

        Args:
            context: QMI context.
            name: Name for this instrument instance.
            transport:  Either a transport string (see create_transport) or a QMI_Transport.
        """
        super().__init__(context, name)
        if isinstance(transport, str):
            self._transport = create_transport(transport)
        elif isinstance(transport, QMI_Transport):
            self._transport = transport

    @rpc_method
    def open(self) -> None:
        """See base class."""
        self._transport.open()
        super().open()

    @rpc_method
    def close(self) -> None:
        """See base class."""
        super().close()
        self._transport.close()

    def _ask(self, command: str, timeout=1) -> str:
        super()._check_is_open()
        bytes_command = "{}\r\n".format(command).encode(encoding='ascii')
        self._transport.write(bytes_command)
        content_line = self._transport.read_until(b'\r\n', timeout)
        done_line = self._transport.read_until(b'\r\n', timeout)
        if done_line.decode('ASCII') != 'Done\r\n':
            raise QMI_InstrumentException('Device did not respond correctly, expected Done message from device.')
        response = content_line.decode('ASCII')
        return response

    @staticmethod
    def _check_channel(channel: int) -> None:
        if not (1 <= channel <= 4):
            raise ValueError("{} is not a valid channel".format(channel))

    @rpc_method
    def get_frequencies(self) -> Tuple[int, ...]:
        """Get the present scrambling frequencies.

        Returns:
            (int, int, int, int) tuple of frequencies [Hz] belonging to channel 1..4 respectively. The returned
            frequencies will be in the range of 0 to 100 Hertz.
        """
        response = self._ask('F?')
        # Device responds with string like "Frequency(Hz): CH1 007  CH2 017  CH3 037  CH4 071", use regex to find all
        # digits for each channel.
        matched_frequencies = re.findall(r'CH. (\d+)', response)
        # convert strings to integers
        frequencies = tuple(map(int, matched_frequencies))
        return frequencies

    @rpc_method
    def set_frequency(self, channel: int, frequency: int) -> None:
        """Set channel to specified frequency.

        Args:
            channel (int): Channel number 1..4.
            frequency (int): Frequency [Hz] in the range of 0 to 100 Hz.
        """
        self._check_channel(channel)

        if not (0 <= frequency <= 100):
            raise ValueError("{} is not a valid channel".format(channel))

        command = 'F{},{}'.format(channel, frequency)
        self._ask(command)

    @rpc_method
    def set_operating_mode_ac(self) -> None:
        """Set operating mode to AC (scrambling)."""
        self._ask('MAC')

    @rpc_method
    def set_operating_mode_dc(self) -> None:
        """Set operating mode to DC (fixed voltage)."""
        self._ask('MDC')

    @rpc_method
    def enable_dc_in_ac_mode(self) -> None:
        """Enable dc voltage in ac mode

        Enables the user to use a fixed dc voltage while the unit is operating in AC mode.
        """
        self._ask('ENVF1')

    @rpc_method
    def disable_dc_in_ac_mode(self) -> None:
        """Disable dc voltage in ac mode

        Disable the user to use a fixed dc voltage while the unit is operating in AC mode.
        """
        self._ask('ENVF0')

    @rpc_method
    def toggle_channel_ac_dc(self, channel: int) -> None:
        """Toggle channel output between AC and DC.

        Notes:
            The device must be in "dc in ac mode" for this operating mode to be effective.

        Args:
            channel (int): Channel number 1..4.
        """
        self._check_channel(channel)
        self._ask('VF{}'.format(channel))

    @rpc_method
    def get_ac_dc_channel_status(self) -> Tuple[str, ...]:
        """Return operating mode of each of the channels.

        Returns:
            (str, str, str, str) tuple of strings, belonging to channel 1..4 respectively, indicating if the channel is
            in 'V' DC mode or 'F' AC mode.
        """
        ac_dc_status = self._ask('VF?')
        channel_status = re.findall(r'CH.-([VF])', ac_dc_status)
        return tuple(channel_status)

    @rpc_method
    def is_in_dc_in_ac_mode(self) -> bool:
        """Device is in dc in ac mode.

        Returns:
            True if the dc in ac mode is enabled, False when the dc in ac mode is disabled.
        """
        ac_dc_status = self._ask('VF?')
        is_dc_in_ac_mode = 'enabled' in ac_dc_status
        return is_dc_in_ac_mode

    @rpc_method
    def get_operating_mode(self) -> str:
        """Get the present operating mode (scrambling or fixed voltage).

        Returns:
            Operating mode string defined as 'AC' or 'DC'. AC (Alternating Current) means that the device is in
            scrambling mode and DC (Direct Current) means the device is in fixed voltage mode.
        """
        response = self._ask('M?')
        words = response.split()
        last_word = words[-1]
        first_two_chars = last_word[:2]
        return first_two_chars

    @rpc_method
    def set_voltage(self, channel: int, voltage: int) -> None:
        """Set channel to the specified voltage.

        Args:
            channel (int): Channel number 1..4.
            voltage (int): Voltage specified in millivolts [mV] that range between -5000 to +5000.
        """
        self._check_channel(channel)

        if not (-5000 <= voltage <= 5000):
            raise ValueError('Provided voltage [mV] is not between -5000 and 5000 mV')

        command = 'V{},{}'.format(channel, voltage)
        self._ask(command)

    @rpc_method
    def set_high(self, channel: int) -> None:
        """Set channel voltage high (maximum).

        Args:
            channel (int): Channel number 1..4.
        """
        self._check_channel(channel)
        command = 'VH{}'.format(channel)
        self._ask(command)

    @rpc_method
    def set_low(self, channel: int) -> None:
        """Set channel voltage low (minimum).

        Args:
            channel: Channel number 1..4.
        """
        self._check_channel(channel)
        command = 'VL{}'.format(channel)
        self._ask(command)

    @rpc_method
    def set_zero(self, channel: int) -> None:
        """Set channel voltage zero volts.

        Args:
            channel: Channel number 1..4.
        """
        self._check_channel(channel)
        command = 'VZ{}'.format(channel)
        self._ask(command)

    @rpc_method
    def get_voltages(self) -> Tuple[int, ...]:
        """Get the output voltages.

        Returns:
            Tuple (int, int, int, int) of voltages [mV] belonging to channel 1..4 respectively. The returned
            voltages will be in the range of -5000 mV to 5000 mV.
        """
        response = self._ask('V?')
        # Device responds with string like "Voltage(mV): CH1 +2200 CH2 +5000 CH3 -1000 CH4 -4000", use regex to find all
        # voltages for each channel.
        matched_frequencies = re.findall(r'CH. (.\d+)', response)
        # convert strings to integers
        frequencies = tuple(map(int, matched_frequencies))
        return frequencies

    @rpc_method
    def get_waveform_type(self) -> str:
        """Get Waveform type (Sine or Triangle)

        This command will get the present waveform type that is being used (Triangle or sine wave).

        Returns:
            A string that indicates either a Sine or a Triangle waveform is used.
        """
        response = self._ask('WF?')
        # Device responds with either "AC Waveform: Sine" or "AC Waveform: Triangle", returning the last word is
        # sufficient here.
        waveform = response.split()[-1]
        return waveform

    @rpc_method
    def save_to_flash(self) -> None:
        """Save current status to flash memory"""
        self._ask('SAVE')

    @rpc_method
    def set_waveform_type_sine(self) -> None:
        """Set waveform type to Sine"""
        self._ask('WF1')

    @rpc_method
    def set_waveform_type_triangle(self) -> None:
        """Set waveform type to Triangle"""
        self._ask('WF2')



#! /usr/bin/env python3

"""Test QMI_Transport functionality."""
import os, sys
import socket
import asyncio
import threading
import time
import unittest.mock
import warnings
import random
import string

import vxi11

import qmi
import qmi.core.exceptions
import qmi.core.transport
from qmi.core.transport import (
    QMI_SocketTransport,
    QMI_UdpTransport,
    QMI_TcpTransport,
    QMI_SerialTransport,
    QMI_Vxi11Transport,
    QMI_UsbTmcTransport
)
from qmi.core.transport import create_transport, list_usbtmc_transports
from qmi.core.instrument import QMI_Instrument
from qmi.core.context import QMI_Context
from qmi.core.rpc import rpc_method
from qmi.utils.context_managers import open_close

from tests.core.pyvisa_stub import ResourceManager, visa_str_1, visa_str_3


class Dummy_USBTMCInstrument(QMI_Instrument):
    """Dummy Instrument driver using USBTMC transport."""

    USB_VENDOR_ID = 0x1313
    USB_PRODUCT_ID = 0x8076

    # Default response timeout in seconds.
    DEFAULT_RESPONSE_TIMEOUT = 5.0

    def __init__(self,
                 context: QMI_Context,
                 name: str,
                 transport: str
                 ) -> None:
        super().__init__(context, name)
        self._timeout = self.DEFAULT_RESPONSE_TIMEOUT
        with unittest.mock.patch("usb.core.find"):
            self._transport = create_transport(transport)

    @rpc_method
    def open(self) -> None:
        self._check_is_closed()
        self._transport.open()
        super().open()

    @rpc_method
    def get_idn(self) -> None:
        return


class TestQmiTransportFactory(unittest.TestCase):
    """Test QMI Transport factory."""

    def setUp(self):
        # Create UDP server socket.
        self.server_sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM, 0)
        self.server_sock.settimeout(10)
        self.server_host, self.server_port = "localhost", 65000
        self.server_sock.bind((self.server_host, self.server_port))

        # Create TCP client socket.
        self.client_sock = socket.socket()
        self.client_sock.settimeout(10)
        self.client_sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
        self.client_sock.bind(("", 0))
        self.client_sock.listen(5)
        (self.client_host, self.client_port) = self.client_sock.getsockname()

    def tearDown(self):
        try:
            self.server_sock.close()
        finally:
            self.client_sock.close()

    @unittest.mock.patch("qmi.core.transport.QMI_SerialTransport")
    def test_factory_serial(self, mock):
        trans = create_transport("serial:/dev/ttyS0:baudrate=9600")
        self.assertIs(trans, mock.return_value)
        mock.assert_called_once_with(device="/dev/ttyS0",
                                     baudrate=9600)

    @unittest.mock.patch("qmi.core.transport.QMI_SerialTransport")
    def test_factory_serial_attributes(self, mock):
        trans = create_transport("serial:COM3:baudrate=115200:bytesize=7:parity=E:stopbits=2:rtscts=True")
        self.assertIs(trans, mock.return_value)
        mock.assert_called_once_with(device="COM3",
                                     baudrate=115200,
                                     bytesize=7,
                                     parity="E",
                                     rtscts=True,
                                     stopbits=2)

    def test_factory_udp(self):
        async def the_call():
            # Async function for not sending the message "too early" so that it won't get lost
            time.sleep(0.1)
            self.server_sock.settimeout(None)
            self.server_sock.sendto(b"aap noot\n", trans._address)
            l = asyncio.get_running_loop()
            l.stop()

        # Create UDP transport.
        with open_close(QMI_UdpTransport("localhost", int(self.server_port) - 1)) as trans:

            # Send some bytes from server to transport.
            loop = asyncio.get_event_loop()
            if loop.is_closed():
                asyncio.set_event_loop(loop)

            loop.run_until_complete(the_call())
            # Receive message through transport.
            data = trans.read_until(b"\n", timeout=1.0)
            self.assertEqual(data, b"aap noot\n")

            if loop.is_running():
                loop.close()

            w_data = b"aap noot\n"
            trans.write(w_data)
            data = trans.read(len(w_data))
            self.assertEqual(data, w_data)

        self.server_sock.close()

    def test_factory_tcp(self):
        # Create TCP transport.
        with open_close(create_transport(f"tcp:localhost:{self.client_port}")) as trans:
            # Accept the connection on client side.
            (client_conn, peer_address) = self.client_sock.accept()
            client_conn.settimeout(1.0)

            # Send some bytes from client to transport.
            client_conn.sendall(b"aap noot\n")

            # Receive message through transport.
            data = trans.read_until(b"\n", timeout=1.0)
            self.assertEqual(data, b"aap noot\n")

        # Check connection closed on server side.
        data = client_conn.recv(100)
        self.assertEqual(data, b"")
        client_conn.close()

    @unittest.mock.patch("qmi.core.transport.QMI_SerialTransport")
    def test_parse_serial_basic(self, mock):
        trans = create_transport("serial:/dev/ttyS0:baudrate=9600")
        self.assertIs(trans, mock.return_value)
        mock.assert_called_once_with(device="/dev/ttyS0",
                                     baudrate=9600)

    @unittest.mock.patch("qmi.core.transport.QMI_SerialTransport")
    def test_parse_serial_attrs(self, mock):
        trans = create_transport("serial:COM3:baudrate=120:bytesize=7:parity=E:stopbits=1.5:rtscts=True")
        self.assertIs(trans, mock.return_value)
        mock.assert_called_once_with(device="COM3",
                                     baudrate=120,
                                     bytesize=7,
                                     parity="E",
                                     stopbits=1.5,
                                     rtscts=True)

    @unittest.mock.patch("qmi.core.transport.QMI_SerialTransport")
    def test_parse_serial_defaults(self, mock):
        trans = create_transport("serial:/dev/ttyUSB0",
                                 {"baudrate": 115200,
                                  "bytesize": 8,
                                  "parity": "N",
                                  "stopbits": 1.0,
                                  "rtscts": False})
        self.assertIs(trans, mock.return_value)
        mock.assert_called_once_with(device="/dev/ttyUSB0",
                                     baudrate=115200,
                                     bytesize=8,
                                     parity="N",
                                     stopbits=1.0,
                                     rtscts=False)

    def test_parse_serial_bad_attrtype(self):
        with self.assertRaises(qmi.core.exceptions.QMI_TransportDescriptorException):
            create_transport("serial:/dev/ttyS0:baudrate=fast")

    def test_parse_serial_bad_booltype(self):
        with self.assertRaises(qmi.core.exceptions.QMI_TransportDescriptorException):
            create_transport("serial:/dev/ttyS0:baudrate=115200:rtscts=0")

    def test_parse_serial_bad_attrname(self):
        with self.assertRaises(qmi.core.exceptions.QMI_TransportDescriptorException):
            create_transport("serial:/dev/ttyS0:spin=up")

    def test_parse_serial_missing_device(self):
        with self.assertRaises(qmi.core.exceptions.QMI_TransportDescriptorException):
            create_transport("serial")

    def test_parse_bad_type(self):
        with self.assertRaises(qmi.core.exceptions.QMI_TransportDescriptorException):
            create_transport("digital:/dev/ttyS0")

    @unittest.mock.patch("qmi.core.transport.QMI_TcpTransport")
    def test_parse_tcp_basic(self, mock):
        trans = create_transport("tcp:localhost:1234")
        self.assertIs(trans, mock.return_value)
        mock.assert_called_once_with(host="localhost", port=1234)

    @unittest.mock.patch("qmi.core.transport.QMI_TcpTransport")
    def test_parse_tcp_ipv4(self, mock):
        trans = create_transport("tcp:192.168.1.222:5000")
        self.assertIs(trans, mock.return_value)
        mock.assert_called_once_with(host="192.168.1.222", port=5000)

    @unittest.mock.patch("qmi.core.transport.QMI_TcpTransport")
    def test_parse_tcp_ipv6(self, mock):
        trans = create_transport("tcp:[2620:0:2d0:200::8]:5000")
        self.assertIs(trans, mock.return_value)
        mock.assert_called_once_with(host="2620:0:2d0:200::8", port=5000)

    @unittest.mock.patch("qmi.core.transport.QMI_TcpTransport")
    def test_parse_tcp_ipv6_range(self, mock):
        trans = create_transport("tcp:[ 2001:db8:1234::/48]:5000")
        self.assertIs(trans, mock.return_value)
        mock.assert_called_once_with(host=" 2001:db8:1234::/48", port=5000)

    @unittest.mock.patch("qmi.core.transport.QMI_TcpTransport")
    def test_parse_tcp_attrs(self, mock):
        trans = create_transport("tcp:localhost:1234:connect_timeout=1")
        self.assertIs(trans, mock.return_value)
        mock.assert_called_once_with(host="localhost", port=1234, connect_timeout=1.0)

    def test_parse_tcp_missing_port(self):
        with self.assertRaises(qmi.core.exceptions.QMI_TransportDescriptorException):
            create_transport("tcp:localhost")

    @unittest.mock.patch("qmi.core.transport.QMI_TcpTransport")
    def test_parse_tcp_default_port(self, mock):
        trans = create_transport("tcp:localhost", {"port": 5000})
        self.assertIs(trans, mock.return_value)
        mock.assert_called_once_with(host="localhost", port=5000)

    @unittest.mock.patch("qmi.core.transport.QMI_TcpTransport")
    def test_parse_tcp_default_port_override(self, mock):
        trans = create_transport("tcp:localhost:21", {"port": 5000})
        self.assertIs(trans, mock.return_value)
        mock.assert_called_once_with(host="localhost", port=21)

    @unittest.mock.patch("qmi.core.transport.QMI_Vxi11Transport")
    def test_parse_vxi11(self, mock):
        trans = create_transport("vxi11:localhost")
        self.assertIs(trans, mock.return_value)
        mock.assert_called_with(host="localhost")


class TestTransportOpenWithDummyInstrument(unittest.TestCase):
    """Test various 'open' cases with a dummy instrument """
    def setUp(self) -> None:
        # Create TCP server socket.
        self.server_sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM, 0)
        self.server_sock.bind(("127.0.0.1", 0))
        (server_host, self.server_port) = self.server_sock.getsockname()
        self.server_sock.listen(5)
        # Create TCP transport.
        self.trans = create_transport("tcp:localhost:%d" % self.server_port)

    def tearDown(self) -> None:
        self.server_sock.close()
        if self.trans._is_open:
            self.trans.close()

    def test_trying_to_open_prototype_directly(self):
        transport = qmi.core.transport.QMI_Transport()
        with self.assertRaises(NotImplementedError):
            transport.open()

        with self.assertRaises(NotImplementedError):
            transport._open_transport()

    def test_trying_to_open_transport_via_private(self):
        expected = None
        res = self.trans._open_transport()
        self.assertEqual(res, expected)
        with self.assertRaises(qmi.core.exceptions.QMI_InvalidOperationException):
            self.trans.discard_read()

    def test_trying_to_open_already_opened_transport(self):
        self.trans.open()
        with self.assertRaises(qmi.core.exceptions.QMI_InvalidOperationException):
            self.trans.open()

    def test_trying_to_act_on_closed_instrument(self):
        with self.assertRaises(qmi.core.exceptions.QMI_InvalidOperationException):
            self.trans.discard_read()


class TestQmiTransportParsing(unittest.TestCase):
    """Test QMI Transport parsing on different OS."""

    def setUp(self):
        # Create TCP server socket.
        self.server_sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM, 0)
        self.server_sock.bind(("127.0.0.1", 0))
        (server_host, self.server_port) = self.server_sock.getsockname()
        self.server_sock.listen(5)

    def tearDown(self):
        self.server_sock.close()

    if "linux" in sys.platform or "darwin" in sys.platform:
        def test_parse_usbtmc(self):
            serialnr = "XYZ"
            vendorid = 0x0699
            productid = 0x3000
            desc = create_transport(f"usbtmc:vendorid={vendorid}:productid={productid}:serialnr={serialnr}")

            self.assertEqual(desc.serialnr, serialnr)
            self.assertEqual(desc.productid, int(productid))
            self.assertEqual(desc.vendorid, int(vendorid))

        def test_parse_usbtmc_non_hex(self):
            serialnr = "XYZ"
            vendorid = 1689
            productid = 12288
            desc = create_transport(f"usbtmc:vendorid={vendorid}:productid={productid}:serialnr={serialnr}")

            self.assertEqual(desc.serialnr, serialnr)
            self.assertEqual(desc.productid, int(productid))
            self.assertEqual(desc.vendorid, int(vendorid))

        def test_parse_usbtmc_defaults(self):
            serialnr = "XYZ"
            vendorid = 0xa1
            productid = 0xb2
            desc = create_transport(f"usbtmc:vendorid={vendorid}:productid={productid}:serialnr={serialnr}")

            self.assertEqual(desc.serialnr, serialnr)
            self.assertEqual(desc.productid, int(productid))
            self.assertEqual(desc.vendorid, int(vendorid))

        def test_parse_gpib_not_for_linux(self):
            primary_addr = 1
            expected_text = "Gpib transport descriptor is for NI GPIB-USB-HS device and Windows-only."

            with self.assertRaises(qmi.core.exceptions.QMI_TransportDescriptorException) as exc:
                create_transport(f"gpib:{primary_addr}")

            self.assertEqual(expected_text, str(exc.exception))

    elif sys.platform.startswith("win") or "msys" in sys.platform:
        # Remove the possibility of import error of pyvisa
        import tests.core.pyvisa_stub
        sys.modules["pyvisa"] = tests.core.pyvisa_stub
        sys.modules["pyvisa.errors"] = tests.core.pyvisa_stub.errors

        def test_parse_usbtmc(self):
            serialnr = "XYZ"
            vendorid = 0x0699
            productid = 0x3000
            transport_str = f"usbtmc:vendorid={vendorid}:productid={productid}:serialnr={serialnr}"
            if os.getenv("LIBUSBPATH"):
                desc = create_transport(transport_str)

            else:
                with unittest.mock.patch("usb.core.find"):
                    desc = create_transport(transport_str)

            self.assertEqual(desc.serialnr, serialnr)
            self.assertEqual(desc.productid, int(productid))
            self.assertEqual(desc.vendorid, int(vendorid))

        def test_parse_usbtmc_non_hex(self):
            serialnr = "XYZ"
            vendorid = 1689
            productid = 12288
            transport_str = f"usbtmc:vendorid={vendorid}:productid={productid}:serialnr={serialnr}"
            if os.getenv("LIBUSBPATH"):
                desc = create_transport(transport_str)

            else:
                with unittest.mock.patch("usb.core.find"):
                    desc = create_transport(transport_str)

            self.assertEqual(desc.serialnr, serialnr)
            self.assertEqual(desc.productid, productid)
            self.assertEqual(desc.vendorid, vendorid)

        def test_parse_usbtmc_defaults(self):
            serialnr = "XYZ"
            vendorid = 0xa1
            productid = 0xb2
            transport_str = f"usbtmc:vendorid={vendorid}:productid={productid}:serialnr={serialnr}"
            if os.getenv("LIBUSBPATH"):
                desc = create_transport(transport_str)

            else:
                with unittest.mock.patch("usb.core.find"):
                    desc = create_transport(transport_str)

            self.assertEqual(desc.serialnr, serialnr)
            self.assertEqual(desc.productid, int(productid))
            self.assertEqual(desc.vendorid, int(vendorid))

        def test_parse_gpib(self):
            board = 0
            primary_addr = 1
            secondary_addr = 2
            timeout = 3.0
            transport_str = f"gpib:board={board}:{primary_addr}:secondary_addr={secondary_addr}:connect_timeout={timeout}"

            if os.getenv("LIBUSBPATH"):
                desc = create_transport(transport_str)

            else:
                with unittest.mock.patch("usb.core.find"):
                    desc = create_transport(transport_str)

            self.assertEqual(desc._board, board)
            self.assertEqual(desc._primary_addr, primary_addr)
            self.assertEqual(desc._secondary_addr, secondary_addr)
            self.assertEqual(desc._connect_timeout, timeout)

        def test_parse_gpib_defaults(self):
            primary_addr = 1
            transport_str = f"gpib:{primary_addr}"
            default_timeout = 30.0

            if os.getenv("LIBUSBPATH"):
                desc = create_transport(transport_str)

            else:
                with unittest.mock.patch("usb.core.find"):
                    desc = create_transport(transport_str)

            self.assertIsNone(desc._board)
            self.assertEqual(primary_addr, desc._primary_addr)
            self.assertIsNone(desc._secondary_addr)
            self.assertEqual(default_timeout, desc._connect_timeout)

    @unittest.mock.patch("qmi.core.transport.QMI_Vxi11Transport")
    def test_parse_vxi11(self, mock):
        trans = create_transport("vxi11:localhost")
        self.assertIs(trans, mock.return_value)
        mock.assert_called_with(host="localhost")


class TestQmiSocketTransportBase(unittest.TestCase):
    """Test QMI_SocketTransportBase class."""

    def setUp(self):
        # Filter out warnings about unclosed sockets
        warnings.filterwarnings("ignore", "unclosed", ResourceWarning)

    def test_host_validation(self):
        """Try to create a base transport with invalid hosts. See that exceptions are raised."""
        with self.assertRaises(qmi.core.exceptions.QMI_TransportDescriptorException):
            QMI_SocketTransport("invalid_hostname", 22)

        with self.assertRaises(qmi.core.exceptions.QMI_TransportDescriptorException):
            QMI_SocketTransport("192.168.1.300", 22)

    def test_port_validation(self):
        """Try to create a base transport with invalid port numbers. See that exceptions are raised."""
        with self.assertRaises(qmi.core.exceptions.QMI_TransportDescriptorException):
            QMI_SocketTransport("localhost", 0)

        with self.assertRaises(qmi.core.exceptions.QMI_TransportDescriptorException):
            QMI_SocketTransport("localhost", 100000)

    def test_write_not_implemented_error(self):
        """Test QMI_SocketTransportBase.write() is not implemented in the base class."""
        # Create UDP transport connected to local server.
        trans = QMI_SocketTransport("localhost", 64500)
        trans.open()

        # Try to send something through the transport.
        with self.assertRaises(NotImplementedError):
            trans.write(b"")


class TestQmiUdpTransport(unittest.TestCase):
    """Test QMI_UdpTransport class."""

    def setUp(self):
        # Filter out warnings about unclosed sockets
        warnings.filterwarnings("ignore", "unclosed", ResourceWarning)

        # Create UDP server socket.
        self.server_sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self.server_host, self.server_port = "localhost", 65000
        self.server_sock.bind((self.server_host, self.server_port))

    def tearDown(self):
        self.server_sock.close()

    def test_cannot_use_qmi_transmission_udp_port(self):
        # Test that port number 35999 is reserved for QMI packets.
        with self.assertRaises(qmi.core.exceptions.QMI_TransportDescriptorException):
            QMI_UdpTransport("localhost", QMI_Context.DEFAULT_UDP_RESPONDER_PORT)

    def test_udp_basic(self):
        """Test basic functionalities of the UDP transport."""
        address = "127.0.0.1" + ':' + str(self.server_port - 1)
        expected_transport_string = f"QMI_UdpTransport(remote={address})"
        # Create UDP transport bound to local server.
        with open_close(QMI_UdpTransport("localhost", int(self.server_port) - 1)) as trans:
            # Check the string name of the instance
            self.assertEqual(expected_transport_string, str(trans))

            # Send some bytes back from server to transport.
            testmsg = b"answer"
            self.server_sock.sendto(testmsg, trans._address)

            # Receive bytes through the transport.
            data = trans.read(len(testmsg), timeout=1.0)
            self.assertEqual(testmsg, data)

            # Try to receive more bytes; wait for timeout.
            with self.assertRaises(qmi.core.exceptions.QMI_TimeoutException):
                trans.read(10, timeout=1.0)

            # Send more bytes back from server to transport.
            self.server_sock.sendto(b"more bytes", trans._address)

            # Receive the first bunch of bytes.
            data = trans.read(5, timeout=1.0)
            self.assertEqual(data, b"more ")

            # Receive the remaining bytes in a separate call.
            # Try to receive more bytes than are available to hit timeout.
            with self.assertRaises(qmi.core.exceptions.QMI_TimeoutException):
                trans.read(10, timeout=1.0)

            # Remaining bytes are still in buffer after failed read above.
            # Receive some of them.
            data = trans.read(3, timeout=1.0)
            self.assertEqual(data, b"byt")

            # Send more bytes from server to transport.
            self.server_sock.sendto(b"first line\nsecond line", trans._address)

            # Receive the first line.
            # It starts with remaining bytes from the previous send.
            data = trans.read_until(b"\n", timeout=1.0)
            self.assertEqual(data, b"esfirst line\n")

            # Receive first word of the second line.
            data = trans.read(7, timeout=None)
            self.assertEqual(data, b"second ")

            # Try to receive a second line; this should hit the timeout.
            with self.assertRaises(qmi.core.exceptions.QMI_TimeoutException):
                trans.read_until(b"\n", timeout=1.0)

            # Send message terminator for the second line.
            self.server_sock.sendto(b"\n", trans._address)

            # Receive second line (note first word was already received).
            data = trans.read_until(b"\n", timeout=1.0)
            self.assertEqual(data, b"line\n")

            # Test with a 2-character terminator.
            self.server_sock.sendto(b"the;last\nline;\n", trans._address)
            data = trans.read_until(b";\n", timeout=1.0)
            self.assertEqual(data, b"the;last\nline;\n")

            # Send some bytes through the transport.
            testmsg = b"aap noot"
            trans.write(testmsg)

            # Receive bytes one-by-one.
            buf = bytearray()
            while len(buf) < len(testmsg):
                data = trans.read_until_timeout(1, timeout=1.0)
                self.assertNotEqual(data, b"")
                buf.extend(data)

            self.assertEqual(buf, testmsg)

    def test_udp_large_packages(self):
        """Test for large package sizes > 4096 bytes, which is regular UDP limit."""
        # Create UDP transport connected to local server.
        with open_close(QMI_UdpTransport("127.0.0.1", self.server_port - 1)) as trans:
            # The transport should fail if sent package is > 4kB, as UDP is not set to handle larger packages.
            bs_size = 5000  # Normal max 2**12 bytes (- headers)
            s = ''.join(random.choices(string.ascii_uppercase + string.digits, k=bs_size))

            # Send some bytes from server to transport.
            self.server_sock.sendto(s.encode(), trans._address)

            try:
                # Try to receive only a part of the bytes (triggers exception).
                with self.assertRaises(qmi.core.exceptions.QMI_RuntimeException):
                    read = trans.read(100, timeout=1.0)  # The `read` will set the size to 4096 in any case

            except AssertionError as ass:
                # Catch this as some servers apparently fragment the message to be max of 4096 bytes, so it does not crash
                if len(read) != 100:
                    raise AssertionError from ass

                trans.discard_read()

            async def the_call():
                # Async function for not sending the message "too early" so that it won't get lost
                time.sleep(0.1)
                self.server_sock.sendto(s.encode(), trans._address)
                l = asyncio.get_running_loop()
                l.stop()

            # Send some bytes from server to transport.
            loop = asyncio.get_event_loop()
            if loop.is_closed():
                asyncio.set_event_loop(loop)

            loop.run_until_complete(the_call())
            try:
                # The same should happen with read_until (triggers exception).
                with self.assertRaises(qmi.core.exceptions.QMI_RuntimeException):
                    trans.read_until(b"\n", timeout=1.0)

            except qmi.core.exceptions.QMI_TimeoutException as tim:
                # Catch this as some servers apparently fragment the message to be max of 4096 bytes, so it does not crash
                if len(trans._read_buffer) != 4096:
                    raise AssertionError from tim

                trans.discard_read()

            if loop.is_running():
                loop.close()

            # Send some bytes from server to transport.
            self.server_sock.sendto(s.encode(), trans._address)

            # With 'read' it is possible to set to read even larger packets. See that it works.
            # The caller does need to know the exact packet size in this case, though.
            data = trans.read(bs_size, timeout=1.0)
            self.assertEqual(data.decode(), s)

    def test_remote_close(self):
        """Test the data is received even after remote is closed."""
        # Create UDP transport connected to local server.
        with open_close(QMI_UdpTransport("127.0.0.1", self.server_port - 1)) as trans:
            # Send some bytes from server to transport.
            self.server_sock.sendto(b"some bytes", trans._address)

            # Close connection on server side.
            self.server_sock.close()

            # Receive bytes through transport.
            data = trans.read(4, timeout=1.0)
            self.assertEqual(data, b"some")

            # Try to receive more bytes than are available (triggers exception).
            with self.assertRaises(qmi.core.exceptions.QMI_TimeoutException):
                trans.read(100, timeout=1.0)

            # Try to receive a line (also triggers exception).
            with self.assertRaises(qmi.core.exceptions.QMI_TimeoutException):
                trans.read_until(b"\n", timeout=0)

            # Receive the remaining bytes.
            data = trans.read(6, timeout=None)
            self.assertEqual(data, b" bytes")

            # Try to read another byte. Should raise exception.
            with self.assertRaises(qmi.core.exceptions.QMI_TimeoutException):
                trans.read(1, timeout=0.0)

    def test_read_until_buffer_already_filled(self):
        """Test that `read_until()` returns immediately if message terminator is already in read buffer."""
        # Create UDP transport bound to local server.
        with open_close(QMI_UdpTransport("localhost", self.server_port - 1)) as trans:
            # Set some bytes with message terminator into read buffer.
            testmsg = b"hello\n"
            trans._read_buffer = testmsg
            # Read and assert
            data = trans.read_until(b"\n", timeout=0.2)
            self.assertEqual(testmsg, data)

    def test_read_until_exceptions(self):
        """Test for exceptions raised with `read_until()` with missing terminal character and too large packet."""
        # Create UDP transport bound to local server.
        with open_close(QMI_UdpTransport("localhost", self.server_port - 1)) as trans:
            # Send some bytes from server to transport.
            testmsg = b"hello there"
            self.server_sock.sendto(testmsg, trans._address)

            # Receive bytes through the transport. But no end character inputted -> get timeout error.
            with self.assertRaises(qmi.core.exceptions.QMI_TimeoutException):
                trans.read_until(b"\n", timeout=0.2)

            # See that the data read until timeout remains in the read buffer.
            self.assertEqual(testmsg, trans._read_buffer)

            # The transport should fail if sent package is > 4kB, as UDP is not set to handle larger packages.
            bs_size = 5000  # Normal max 2**12 bytes (- headers)
            s = ''.join(random.choices(string.ascii_uppercase + string.digits, k=bs_size))

            # Send some bytes from server to transport.
            self.server_sock.sendto(s.encode(), trans._address)

            try:
                # Receive bytes through the transport. But packet too large.
                with self.assertRaises(qmi.core.exceptions.QMI_RuntimeException):
                    trans.read_until(b"\n", timeout=0.2)

            except qmi.core.exceptions.QMI_TimeoutException as tim:
                # Catch this as some servers apparently fragment the message to be max of 4096 bytes, so it does not crash
                if len(trans._read_buffer) != 4107:
                    raise AssertionError from tim

    def test_read_until_timeout(self):
        """Test `read_until_timeout` until timeout."""
        # Create UDP transport connected to local server.
        with open_close(QMI_UdpTransport("localhost", self.server_port - 1)) as trans:
            # Send some bytes from server to transport.
            testmsg = b"hello there"
            self.server_sock.sendto(testmsg, trans._address)

            # Receive bytes through the transport.
            data = trans.read_until_timeout(len(testmsg), timeout=1.0)
            self.assertEqual(data, testmsg)

            # Try to receive more bytes; wait for timeout.
            data = trans.read_until_timeout(10, timeout=1.0)
            self.assertEqual(data, b"")

            # Send more bytes back from server to transport.
            self.server_sock.sendto(b"more bytes", trans._address)

            # Receive the first bunch of bytes.
            data = trans.read_until_timeout(5, timeout=1.0)
            self.assertEqual(data, b"more ")

            # Try to receive more bytes than are available to get partial data
            # after timeout.
            data = trans.read_until_timeout(10, timeout=1.0)
            self.assertEqual(data, b"bytes")

            # Send more bytes from server to transport.
            self.server_sock.sendto(b"last data", trans._address)

            # Close connection on server side.
            self.server_sock.close()

            # Receive bytes through transport.
            data = trans.read_until_timeout(4, timeout=1.0)
            self.assertEqual(data, b"last")

            # Receive remaining bytes (results in partial answer).
            data = trans.read_until_timeout(100, timeout=1.0)
            self.assertEqual(data, b" data")

            # Try to read more bytes. Should raise timeout exception in UDP.
            with self.assertRaises(qmi.core.exceptions.QMI_TimeoutException):
                trans.read(100, timeout=1.0)

    def test_discard_read(self):
        """Test discarding read data."""
        # Create UDP transport connected to local server.
        with open_close(QMI_UdpTransport("localhost", self.server_port + 1)) as trans:
            # Send some bytes from server to transport.
            testmsg = b"hello there"
            self.server_sock.sendto(testmsg, trans._address)

            # Discard bytes through the transport.
            trans.discard_read()
            # Try to read, which should now except as the input buffer was discarded
            with self.assertRaises(qmi.core.exceptions.QMI_TimeoutException):
                trans.read(1, 0.0)

            # The transport should fail if sent package is > 4kB, as UDP is not set to handle larger packages.
            bs_size = 5000  # Normal max 2**12 bytes (- headers)
            s = ''.join(random.choices(string.ascii_uppercase + string.digits, k=bs_size))

            # Send some bytes from server to transport.
            self.server_sock.sendto(s.encode(), trans._address)

            # Try to discard. This should activate the OSError except catch.
            trans.discard_read()

            # Try to read, which should now except as the input buffer was discarded
            with self.assertRaises(qmi.core.exceptions.QMI_TimeoutException):
                trans.read(1, 0.0)

    def test_client_non_existing(self):
        """Test that non-existing UDP client raises an error."""
        # Try to create a server for non-existing client host name.
        trans = QMI_UdpTransport("imaginary-host", 5123)
        with self.assertRaises(OSError):
            trans.open()

    def test_opening_twice_error(self):
        """The same UDP host and port should not be possible to open twice."""
        # Create UDP transport connected to local server.
        with open_close(QMI_UdpTransport("localhost", self.server_port - 2)):
            # Try to create a second one with same number. Should raise an error.
            trans2 = QMI_UdpTransport("localhost", self.server_port - 2)
            with self.assertRaises((OSError, PermissionError)):
                trans2.open()

    def test_udp_async(self):
        """Test asynchronous data sending and receiving works, also with partial reads."""
        # Create UDP transport connected to local server.
        with open_close(QMI_UdpTransport("localhost", self.server_port - 1)) as trans:
            # Create a separate thread which sends data from the server side.
            def thread_main():
                for i in range(4):
                    time.sleep(0.5)
                    self.server_sock.sendto("tick {}".format(i).encode("ascii"), trans._address)
                    time.sleep(0.5)
                    self.server_sock.sendto("tock {}\n".format(i).encode("ascii"), trans._address)

            thread = threading.Thread(target=thread_main, name="test_transport_thread")
            thread.start()

            try:
                # Receive data through the transport (after ~ 0.5 second delay).
                data = trans.read(6, timeout=1.0)
                self.assertEqual(data, b"tick 0")

                # Receive terminated message through the transport (after delay).
                data = trans.read_until(b"\n", timeout=1.0)
                self.assertEqual(data, b"tock 0\n")

                # Receive data which arrives in 3 chunks.
                data = trans.read(15, timeout=2.0)
                self.assertEqual(data, b"tick 1tock 1\nti")

                # Receive terminated message which arrives in chunks.
                data = trans.read_until(b"\n", timeout=1.0)
                self.assertEqual(data, b"ck 2tock 2\n")

                # Receive partial data after timeout.
                data = trans.read_until_timeout(100, timeout=0.6)
                self.assertEqual(data, b"tick 3")

                # Check no further data available yet.
                with self.assertRaises(qmi.core.exceptions.QMI_TimeoutException):
                    trans.read_until(b"\n", timeout=0)

                # Wait until data arrives.
                data = trans.read_until(b"\n", timeout=1.0)
                self.assertEqual(data, b"tock 3\n")

                # Check no further data available.
                data = trans.read_until_timeout(1, timeout=0)
                self.assertEqual(data, b"")

            finally:
                # Clean up background thread.
                thread.join()
                # Close server side.
                self.server_sock.close()


class TestQmiTcpTransport(unittest.TestCase):
    """Test QMI_TcpTransport class."""

    def setUp(self):
        # Filter out warnings about unclosed sockets
        warnings.filterwarnings("ignore", "unclosed", ResourceWarning)
        # Create dummy non-listening TCP port.
        self.dummy_sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM, 0)
        self.dummy_sock.bind(("127.0.0.1", 0))
        (dummy_host, self.dummy_port) = self.dummy_sock.getsockname()

        # Create TCP server socket.
        self.server_sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM, 0)
        self.server_sock.bind(("127.0.0.1", 0))
        (server_host, self.server_port) = self.server_sock.getsockname()
        self.server_sock.listen(5)

    def tearDown(self):
        self.dummy_sock.close()
        self.server_sock.close()

    def test_tcp_basic(self):
        """Test basic functionalities of TCP transport work."""
        address = "127.0.0.1" + ':' + str(self.server_port)
        expected_transport_string = f"QMI_TcpTransport(remote={address})"
        # Create TCP transport connected to local server.
        with open_close(QMI_TcpTransport("localhost", self.server_port, connect_timeout=1)) as trans:
            # Check the string name of the instance
            self.assertEqual(expected_transport_string, str(trans))

            # Accept the connection on server side.
            (server_conn, peer_address) = self.server_sock.accept()
            server_conn.settimeout(1.0)

            # Send some bytes through the transport.
            testmsg = b"aap noot"
            trans.write(testmsg)

            # Receive bytes at the server side.
            buf = bytearray()
            while len(buf) < len(testmsg):
                data = server_conn.recv(100)
                self.assertNotEqual(data, b"")
                buf.extend(data)
            self.assertEqual(buf, testmsg)

            # Send some bytes back from server to transport.
            testmsg = b"answer"
            server_conn.sendall(testmsg)

            # Receive bytes through the transport.
            data = trans.read(len(testmsg), timeout=1.0)
            self.assertEqual(data, testmsg)

            # Try to receive more bytes; wait for timeout.
            with self.assertRaises(qmi.core.exceptions.QMI_TimeoutException):
                trans.read(10, timeout=1.0)

            # Send more bytes back from server to transport.
            server_conn.sendall(b"more bytes")

            # Receive the first bunch of bytes.
            data = trans.read(5, timeout=1.0)
            self.assertEqual(data, b"more ")

            # Receive the remaining bytes in a separate call.
            # Try to receive more bytes than are available to hit timeout.
            with self.assertRaises(qmi.core.exceptions.QMI_TimeoutException):
                trans.read(10, timeout=1.0)

            # Remaining bytes are still in buffer after failed read above.
            # Receive some of them.
            data = trans.read(3, timeout=1.0)
            self.assertEqual(data, b"byt")

            # Send more bytes from server to transport.
            server_conn.sendall(b"first line\nsecond line")

            # Receive the first line.
            # It starts with remaining bytes from the previous send.
            data = trans.read_until(b"\n", timeout=1.0)
            self.assertEqual(data, b"esfirst line\n")

            # Receive first word of the second line.
            data = trans.read(7, timeout=None)
            self.assertEqual(data, b"second ")

            # Try to receive a second line; this should hit the timeout.
            with self.assertRaises(qmi.core.exceptions.QMI_TimeoutException):
                trans.read_until(b"\n", timeout=1.0)

            # Send message terminator for the second line.
            server_conn.sendall(b"\n")

            # Receive second line (note first word was already received).
            data = trans.read_until(b"\n", timeout=1.0)
            self.assertEqual(data, b"line\n")

            # Test with a 2-character terminator.
            server_conn.sendall(b"the;last\nline;\n")
            data = trans.read_until(b";\n", timeout=1.0)
            self.assertEqual(data, b"the;last\nline;\n")

        # Check upstream channel closed.
        data = server_conn.recv(100)
        self.assertEqual(data, b"")

        # Close server-side socket.
        server_conn.close()

    def test_remote_close(self):
        """Test data is received also after remote is closed"""
        # Create TCP transport connected to local server.
        with open_close(QMI_TcpTransport("127.0.0.1", self.server_port, connect_timeout=1)) as trans:
            # Accept the connection on server side.
            (server_conn, peer_address) = self.server_sock.accept()
            server_conn.settimeout(1.0)

            # Send some bytes from server to transport.
            server_conn.sendall(b"some bytes")

            # Close connection on server side.
            server_conn.close()

            # Receive bytes through transport.
            data = trans.read(4, timeout=1.0)
            self.assertEqual(data, b"some")

            # Try to receive more bytes than are available (triggers exception).
            with self.assertRaises(qmi.core.exceptions.QMI_EndOfInputException):
                trans.read(100, timeout=1.0)

            # Try to receive a line (also triggers exception).
            with self.assertRaises(qmi.core.exceptions.QMI_EndOfInputException):
                trans.read_until(b"\n", timeout=None)

            # Receive the remaining bytes.
            data = trans.read(6, timeout=None)
            self.assertEqual(data, b" bytes")

            # Try to read another byte. Should raise exception.
            with self.assertRaises(qmi.core.exceptions.QMI_EndOfInputException):
                trans.read(1, timeout=None)

    def test_read_until_buffer_already_filled(self):
        """Test that `read_until()` returns immediately if message terminator is already in read buffer."""
        # Create TCP transport connected to local server.
        with open_close(QMI_TcpTransport("localhost", self.server_port, connect_timeout=1)) as trans:
            # Set some bytes with message terminator into read buffer.
            testmsg = b"hello\n"
            trans._read_buffer = testmsg
            # Read and assert
            data = trans.read_until(b"\n", timeout=0.2)
            self.assertEqual(testmsg, data)

    def test_read_until_timeout(self):
        """Test `read_until_timeout` until timeout."""
        # Create TCP transport connected to local server.
        with open_close(QMI_TcpTransport("localhost", self.server_port, connect_timeout=1)) as trans:
            # Accept the connection on server side.
            (server_conn, peer_address) = self.server_sock.accept()
            server_conn.settimeout(1.0)

            # Send some bytes from server to transport.
            testmsg = b"hello there"
            server_conn.sendall(testmsg)

            # Receive bytes through the transport.
            data = trans.read_until_timeout(len(testmsg), timeout=1.0)
            self.assertEqual(data, testmsg)

            # Try to receive more bytes; wait for timeout.
            data = trans.read_until_timeout(10, timeout=1.0)
            self.assertEqual(data, b"")

            # Send more bytes back from server to transport.
            server_conn.sendall(b"more bytes")

            # Receive the first bunch of bytes.
            data = trans.read_until_timeout(5, timeout=1.0)
            self.assertEqual(data, b"more ")

            # Try to receive more bytes than are available to get partial data
            # after timeout.
            data = trans.read_until_timeout(10, timeout=1.0)
            self.assertEqual(data, b"bytes")

            # Send more bytes from server to transport.
            server_conn.sendall(b"last data")

            # Close connection on server side.
            server_conn.close()

            # Receive bytes through transport.
            data = trans.read_until_timeout(4, timeout=1.0)
            self.assertEqual(data, b"last")

            # Receive remaining bytes (results in partial answer).
            data = trans.read_until_timeout(100, timeout=1.0)
            self.assertEqual(data, b" data")

            # Try to read more bytes. Should raise exception.
            with self.assertRaises(qmi.core.exceptions.QMI_EndOfInputException):
                trans.read(100, timeout=1.0)

    def test_connect_failure(self):
        """Test connecting to non-existing or closed or unroutable host fails."""
        # Try to connect to non-existing host name.
        with self.assertRaises(OSError):
            trans = QMI_TcpTransport("imaginary-host", 5123, connect_timeout=1)
            trans.open()

        # Try to connect to closed TCP port.
        # This may trigger connection refused, or timeout if the port is firewalled.
        with self.assertRaises((OSError, qmi.core.exceptions.QMI_TimeoutException)):
            trans = QMI_TcpTransport("localhost", self.dummy_port, connect_timeout=1)
            trans.open()

        # Try to connect to unroutable IP address.
        with self.assertRaises(Exception):
            trans = QMI_TcpTransport("192.168.231.11", 5123, connect_timeout=1)
            trans.open()

    def test_tcp_async(self):
        """Test asynchronous sending and receiving data works, with partial data."""
        # Create TCP transport connected to local server.
        with open_close(QMI_TcpTransport("localhost", self.server_port, connect_timeout=1)) as trans:
            # Accept the connection on server side.
            (server_conn, peer_address) = self.server_sock.accept()
            server_conn.settimeout(1.0)

            # Create a separate thread which sends data from the server side.
            def thread_main():
                for i in range(4):
                    time.sleep(0.5)
                    server_conn.sendall("tick {}".format(i).encode("ascii"))
                    time.sleep(0.5)
                    server_conn.sendall("tock {}\n".format(i).encode("ascii"))

            thread = threading.Thread(target=thread_main, name="test_transport_thread")
            thread.start()

            try:

                # Receive data through the transport (after ~ 0.5 second delay).
                data = trans.read(6, timeout=1.0)
                self.assertEqual(data, b"tick 0")

                # Receive terminated message through the transport (after delay).
                data = trans.read_until(b"\n", timeout=1.0)
                self.assertEqual(data, b"tock 0\n")

                # Receive data which arrives in 3 chunks.
                data = trans.read(15, timeout=2.0)
                self.assertEqual(data, b"tick 1tock 1\nti")

                # Receive terminated message which arrives in chunks.
                data = trans.read_until(b"\n", timeout=1.0)
                self.assertEqual(data, b"ck 2tock 2\n")

                # Receive partial data after timeout.
                data = trans.read_until_timeout(100, timeout=0.6)
                self.assertEqual(data, b"tick 3")

                # Check no further data available yet.
                with self.assertRaises(qmi.core.exceptions.QMI_TimeoutException):
                    trans.read_until(b"\n", timeout=0)

                # Wait until data arrives.
                data = trans.read_until(b"\n", timeout=1.0)
                self.assertEqual(data, b"tock 3\n")

                # Check no further data available.
                data = trans.read_until_timeout(1, timeout=0)
                self.assertEqual(data, b"")

            finally:
                # Clean up background thread.
                thread.join()
                # Close server side.
                server_conn.close()

    def test_invalid_host(self):
        with self.assertRaises(qmi.core.exceptions.QMI_TransportDescriptorException):
            QMI_TcpTransport("invalid_hostname", 22)
        with self.assertRaises(qmi.core.exceptions.QMI_TransportDescriptorException):
            QMI_TcpTransport("192.168.1.300", 22)


class TestQmiSerialTransportInit(unittest.TestCase):

    def test_correct_construction(self):
        QMI_SerialTransport("COM1", 9600)
        QMI_SerialTransport("/valid/path", 9600)
        QMI_SerialTransport("COM3", 120, bytesize=7, parity="E", stopbits=1.5)
        QMI_SerialTransport("/dev/ttyUSB0", 115200, bytesize=8, parity="N", stopbits=1.0)
        QMI_SerialTransport("/dev/ttyUSB0", 115200, bytesize=8, parity="N", stopbits=1.0, rtscts=True)

    def test_invalid_device_description(self):
        with self.assertRaises(qmi.core.exceptions.QMI_TransportDescriptorException):
            QMI_SerialTransport("invalid", 9600)

    def test_invalid_baudrate(self):
        with self.assertRaises(qmi.core.exceptions.QMI_TransportDescriptorException):
            QMI_SerialTransport("/valid/path", 0)

    def test_invalid_bytesize(self):
        with self.assertRaises(qmi.core.exceptions.QMI_TransportDescriptorException):
            QMI_SerialTransport("/valid/path", 9600, bytesize=4)
        with self.assertRaises(qmi.core.exceptions.QMI_TransportDescriptorException):
            QMI_SerialTransport("/valid/path", 9600, bytesize=9)

    def test_invalid_parity(self):
        with self.assertRaises(qmi.core.exceptions.QMI_TransportDescriptorException):
            QMI_SerialTransport("/valid/path", 9600, parity="X")
        with self.assertRaises(qmi.core.exceptions.QMI_TransportDescriptorException):
            QMI_SerialTransport("/valid/path", 9600, parity="FOO")

    def test_invalid_stopbits(self):
        with self.assertRaises(qmi.core.exceptions.QMI_TransportDescriptorException):
            QMI_SerialTransport("/valid/path", 9600, stopbits=1.4)

    def test_invalid_rtscts(self):
        with self.assertRaises(qmi.core.exceptions.QMI_TransportDescriptorException):
            QMI_SerialTransport("/valid/path", 9600, rtscts=3.14)

    @unittest.mock.patch("qmi.core.transport.serial.Serial", autospec=True)
    def test_correct_open_close(self, mocked_serial):
        device = "/dev/mydevice"
        baudrate = 9600
        transport = QMI_SerialTransport(device, baudrate)  # use defaults for other arguments
        transport.open()
        transport.close()

        mocked_serial.assert_called_once_with(
            device, baudrate=baudrate,
            bytesize=8, parity="N", stopbits=1.0, rtscts=False,
            timeout=QMI_SerialTransport.SERIAL_READ_TIMEOUT
        )
        mocked_serial.return_value.close.assert_called_once()


class TestQmiSerialTransportMethods(unittest.TestCase):

    def setUp(self):
        patcher = unittest.mock.patch("qmi.core.transport.serial.Serial", autospec=True)
        self.serial = patcher.start().return_value
        self.addCleanup(patcher.stop)

        self.transport = QMI_SerialTransport("/dev/mydevice", 9600)
        self.transport.open()

    def tearDown(self):
        self.transport.close()
        self.serial.close.assert_called_once()

    def test_write(self):
        data = b"hello, world!"
        self.transport.write(data)
        self.serial.write.assert_called_once_with(data)

    def test_read_blocking_immediate(self):
        data = b"hello, world!"
        nbytes = len(data)
        self.serial.read.return_value = data

        recv = self.transport.read(nbytes, timeout=None)

        self.assertEqual(recv, data)
        self.serial.read.assert_called_once_with(nbytes)

    def test_read_blocking_sequential(self):
        data = b"hello, world!"
        nbytes = len(data)
        self.serial.read.side_effect = [bytearray([d]) for d in data]

        recv = self.transport.read(nbytes, timeout=None)

        self.assertEqual(recv, data)

    def test_read_blocking_timeout(self):
        self.serial.read.return_value = bytearray([])

        with self.assertRaises(qmi.core.exceptions.QMI_TimeoutException):
            self.transport.read(1, timeout=0.1)

    def test_read_nonblocking(self):
        data = b"hello, world!"
        nbytes = len(data)
        self.serial.in_waiting = nbytes
        self.serial.read.return_value = data

        recv = self.transport.read(nbytes, timeout=0)

        self.assertEqual(recv, data)
        self.serial.read.assert_called_once_with(nbytes)

    def test_read_timeout_retry(self):
        data = b"hello, world!"
        nbytes = len(data)
        nbytes_partial = 5

        # Helper function that only returns part of the data array on the first call and then stalls.
        def _read(n):
            _read.calls += 1
            if _read.calls == 1:
                return data[:nbytes_partial]
            else:
                return bytearray([])

        _read.calls = 0
        self.serial.read.side_effect = _read

        try:
            self.transport.read(nbytes, timeout=0.1)
        except qmi.core.exceptions.QMI_TimeoutException:
            pass  # deliberately timeout

        # The partial read should be in the buffer (no more calls to read())
        self.serial.read.reset_mock()
        recv = self.transport.read(nbytes_partial, None)
        self.assertEqual(recv, data[:nbytes_partial])
        self.serial.read.assert_not_called()

    def test_read_until_immediate(self):
        data = b"hello, world!"

        self.serial.in_waiting = len(data)
        self.serial.read.return_value = data

        recv = self.transport.read_until(b',', timeout=None)
        self.assertEqual(recv, b"hello,")

    def test_read_until_sequential(self):
        data = b"hello, world!"
        nbytes_partial = 3

        self.serial.in_waiting = nbytes_partial
        self.serial.read.side_effect = [data[:nbytes_partial]] + [bytearray([d]) for d in data[nbytes_partial:]]

        recv = self.transport.read_until(b',', timeout=None)
        self.assertEqual(recv, b"hello,")

    def test_read_until_with_timeout(self):
        self.serial.in_waiting = 1
        self.serial.read.return_value = b'a'

        with self.assertRaises(qmi.core.exceptions.QMI_TimeoutException):
            self.transport.read_until(b',', timeout=0.1)

    def test_read_until_timeout(self):
        data = b"hello, world!"
        nbytes = len(data)
        nbytes_partial = 5

        # Helper function that only returns part of the data array on the first call and then stalls.
        def _read(n):
            _read.calls += 1
            if _read.calls == 1:
                return data[:nbytes_partial]
            else:
                return bytearray([])

        _read.calls = 0
        self.serial.read.side_effect = _read

        recv = self.transport.read_until_timeout(nbytes, timeout=0.1)
        self.assertEqual(recv, data[:nbytes_partial])

    def test_read_until_multiple(self):
        data = b"hello;world;"
        nbytes = len(data)
        self.serial.in_waiting = nbytes
        self.serial.read.return_value = data

        # Read from serial.
        recv1 = self.transport.read_until(b';', timeout=None)
        self.assertEqual(recv1, b"hello;")

        # Read from buffer.
        recv2 = self.transport.read_until(b';', timeout=None)
        self.assertEqual(recv2, b"world;")

        self.serial.read.assert_called_once_with(nbytes)

    def test_discard_read(self):
        self.transport.discard_read()
        self.serial.reset_input_buffer.assert_called_once()

    def test_read_with_discard(self):
        data = b"hello;world;"
        nbytes = len(data)
        self.serial.in_waiting = nbytes
        self.serial.read.return_value = data

        # Read from serial.
        recv1 = self.transport.read_until(b';', timeout=None)
        self.assertEqual(recv1, b"hello;")

        # Discard buffer.
        self.transport.discard_read()

        # Read will again go via serial, because buffer is empty.
        recv2 = self.transport.read_until(b';', timeout=None)
        self.assertEqual(recv2, b"hello;")

        self.serial.read.assert_has_calls([unittest.mock.call(nbytes), unittest.mock.call(nbytes)])
        self.serial.reset_input_buffer.assert_called_once()


class TestQmiUsbTmcTransport(unittest.TestCase):

    def test_correct_construction(self):
        vendorid = 0x0699
        productid = 0x3000
        serialnr = "XYZ"
        exp_trnsprt = f"QMI_UsbTmcTransport 0x{vendorid:04x}:0x{productid:04x} ({serialnr})"
        # Make two transports, one with hex input and other with int input
        trnsprt_1 = QMI_UsbTmcTransport(vendorid, productid, "XYZ")
        trnsprt_2 = QMI_UsbTmcTransport(1689, 12288, "XYZ")  # Same values but in int
        # Assert that transport string should be always formatted with hex numbers
        self.assertEqual(exp_trnsprt, str(trnsprt_1))
        self.assertEqual(exp_trnsprt, str(trnsprt_2))

    def test_invalid_vendor_id(self):
        """Test _validate_vendor_id method."""
        with self.assertRaises(qmi.core.exceptions.QMI_TransportDescriptorException):
            QMI_UsbTmcTransport(0xfcafe, 0xbebe, "bar")

        with self.assertRaises(qmi.core.exceptions.QMI_TransportDescriptorException):
            QMI_UsbTmcTransport(-1, 0xbebe, "foo")

    def test_invalid_product_id(self):
        """Test _validate_product_id method."""
        with self.assertRaises(qmi.core.exceptions.QMI_TransportDescriptorException):
            QMI_UsbTmcTransport(0xbebe, 0xfcafe, "bar")

        with self.assertRaises(qmi.core.exceptions.QMI_TransportDescriptorException):
            QMI_UsbTmcTransport(0xbebe, -1, "foo")

    def test_not_implemented_methods(self):
        """Test not implemented functions raise NotImplementedError"""
        transport = QMI_UsbTmcTransport(1689, 12288, "XYZ")
        with self.assertRaises(NotImplementedError):
            transport.write(b"no_data")

        with self.assertRaises(NotImplementedError):
            transport._read_message(None)

        with self.assertRaises(NotImplementedError):
            transport.list_resources()


class TestQmiVxi11TransportInit(unittest.TestCase):

    @unittest.mock.patch("qmi.core.transport.vxi11.Instrument")
    def test_init(self, mock: unittest.mock.Mock):
        """The init of the transport only saves the arguments, the instrument is created only at open. """
        QMI_Vxi11Transport("localhost")
        mock.assert_not_called()

    def test_init_invalid_host(self):
        """The init will raise an exception due to invalid host. """
        with self.assertRaises(qmi.core.exceptions.QMI_TransportDescriptorException):
            QMI_Vxi11Transport("-invalid-host")

    @unittest.mock.patch("qmi.core.transport.vxi11.Instrument")
    def test_open(self, mock: unittest.mock.Mock):
        """Test open of device.

        Expecting:
            The instrument to be made with host.
            The instrument to be opened.
        """
        host = "localhost"
        instr = QMI_Vxi11Transport(host)
        instr.open()
        mock.assert_called_once_with(host)
        mock().open.assert_called_once()

    @unittest.mock.patch("qmi.core.transport.vxi11.Instrument")
    def test_open_vxi11_error(self, mock: unittest.mock.Mock):
        """Test open with VXI11 error.

        Expecting:
            QMI_InstrumentException has wrapped the internal error.
        """
        instr = QMI_Vxi11Transport("localhost")
        mock().open.side_effect = vxi11.vxi11.Vxi11Exception(note="Test error!")
        with self.assertRaises(qmi.core.exceptions.QMI_InstrumentException):
            instr.open()

    @unittest.mock.patch("qmi.core.transport.vxi11.Instrument")
    def test_close(self, mock: unittest.mock.Mock):
        """Test close of the transport.

        Expecting:
            The underlying VXI11 instrument to be closed.
        """
        instr = QMI_Vxi11Transport("localhost")
        instr.open()
        instr.close()
        mock().close.assert_called_once()

    @unittest.mock.patch("qmi.core.transport.vxi11.Instrument")
    def test_close_vxi11_error(self, mock: unittest.mock.Mock):
        """Test close with VXI11 error.

        Expecting:
            QMI_InstrumentException has wrapped the internal error.
        """
        instr = QMI_Vxi11Transport("localhost")
        instr.open()
        mock().close.side_effect = vxi11.vxi11.Vxi11Exception(note="Test error!")
        with self.assertRaises(qmi.core.exceptions.QMI_InstrumentException):
            instr.close()


class TestQmiVxi11TransportMethods(unittest.TestCase):

    def setUp(self):
        patcher = unittest.mock.patch("qmi.core.transport.vxi11.Instrument")
        self.mock = patcher.start()
        self.addCleanup(patcher.stop)

        self.instr = QMI_Vxi11Transport("localhost")
        self.instr.open()

    def test_write(self):
        """Test writing to VXI11 instrument

        Expecting:
            Bytes has been passed untouched to vxi11 instrument.
        """
        self.instr.write(unittest.mock.sentinel.data)
        self.mock().write_raw.assert_called_once_with(unittest.mock.sentinel.data)

    def test_write_error(self):
        """Test exception during writing. """
        self.mock().write_raw.side_effect = vxi11.vxi11.Vxi11Exception(note="Test error!")
        with self.assertRaises(qmi.core.exceptions.QMI_InstrumentException):
            self.instr.write(unittest.mock.sentinel.data)

    def test_read(self):
        """Test reading from VXI11 instrument.

        Expecting:
            Bytes read from VXI11 instrument has been passed untouched.
        """
        # Arrange
        expected_read = b"a1b2c3d4"
        nbytes = len(expected_read)
        # Making data to be a bytearray, it gets all read in one go, not byte-per-byte.
        self.mock().read_raw.side_effect = [bytearray(expected_read)]
        # Act
        data = self.instr.read(nbytes)
        # Assert
        self.assertEqual(expected_read, data)
        self.mock().read_raw.assert_called_once_with(self.instr._instr.max_recv_size)

    def test_read_buffer_already_filled(self):
        """Test reading from VXI11 instrument with already filled buffer.

        Expecting:
            Data is returned from the already filled buffer.
            read_raw() is not called.
        """
        expected_read = b"a1b2c3d4"
        nbytes = len(expected_read)
        self.instr._read_buffer = expected_read + b"\n"
        data = self.instr.read(nbytes)
        # Assert
        self.mock().read_raw.assert_not_called()
        self.assertEqual(expected_read, data)

    def test_read_timeout(self):
        """Test timeout exception during read.

        Expecting:
            Timeout is set.
            QMI_TimeoutException is raised.
            Timeout is reverted to default.
        """
        self.mock().timeout = 10.0  # default timeout
        self.mock().read_raw.side_effect = vxi11.vxi11.Vxi11Exception(err=15, note="Test error!")
        with self.assertRaises(qmi.core.exceptions.QMI_TimeoutException):
            self.instr.read(nbytes=8, timeout=0.5)

        # Assert that the timeout is restored
        self.assertAlmostEqual(self.mock().timeout, 10.0)
        self.mock().read_raw.assert_called_once_with(self.instr._instr.max_recv_size)

    def test_read_error(self):
        """Test exception during read. See that the _read_buffer contains data until exception.

        Expecting:
            A few bytes are read.
            QMI_EndOfInputException is raised.
            The read bytes remain in the buffer and are not discarded.
        """
        # Arrange
        test_string = b"test"
        self.mock().read_raw.side_effect = [bytearray(test_string)] +\
           [vxi11.vxi11.Vxi11Exception(note="Test error!")]
        with self.assertRaises(qmi.core.exceptions.QMI_EndOfInputException):
            self.instr.read(8, 0.1)  # Try to read 8 bytes

        self.assertEqual(test_string, self.instr._read_buffer)
        self.assertEqual(2, self.mock().read_raw.call_count)

    def test_read_until(self):
        """Test read_until message terminator command.

        Expecting:
            Default timeout isn't changed after call
            VXI11 call to read data from instrument
            Data mocked is returned
        """
        self.mock().timeout = 10.0  # default timeout
        test_string = "test\n".encode("utf-8")
        self.mock().read_raw.side_effect = [chr(c).encode() for c in test_string]
        data = self.instr.read_until(message_terminator="\n".encode("utf-8"), timeout=1.5)
        self.assertAlmostEqual(self.mock().timeout, 10.0)
        self.assertEqual(len(test_string), self.mock().read_raw.call_count)
        self.assertEqual(test_string, data)

    def test_read_until_buffer_already_filled(self):
        """Test read_until with buffer already filled.

        Expecting:
            Data is returned from buffer.
            read_raw() is not called.
        """
        test_string = "test\n".encode("utf-8")
        self.instr._read_buffer = test_string
        data = self.instr.read_until(message_terminator="\n".encode("utf-8"))
        # Assert
        self.mock().read_raw.assert_not_called()
        self.assertEqual(test_string.strip(), data)

    def test_read_until_invalid_term_char(self):
        """Test that a too long terminator raises an error. """
        with self.assertRaises(qmi.core.exceptions.QMI_InstrumentException):
            _ = self.instr.read_until(message_terminator=bytes("error", "utf-8"), timeout=None)

    def test_read_until_error(self):
        """Tests whether instrument error raises a QMI instrument exception. """
        self.mock().read_raw.side_effect = vxi11.vxi11.Vxi11Exception(note="Test error!")
        with self.assertRaises(qmi.core.exceptions.QMI_InstrumentException):
            _ = self.instr.read_until(message_terminator=bytes("\n", "utf-8"), timeout=None)

    def test_read_until_timeout_exception(self):
        """Tests whether an instrument timeout raises a QMI timeout exception. """
        self.mock().read_raw.side_effect = vxi11.vxi11.Vxi11Exception(err=15, note="Test error!")
        with self.assertRaises(qmi.core.exceptions.QMI_TimeoutException):
            _ = self.instr.read_until(message_terminator=bytes("\n", "utf-8"), timeout=None)

    def test_read_until_missing_term_char(self):
        """Tests whether the function catches a missing terminator from instrument. """
        self.mock().read_raw.side_effect = [b"test", vxi11.vxi11.Vxi11Exception(err=15, note="Test error!")]
        with self.assertRaises(qmi.core.exceptions.QMI_TimeoutException):
            _ = self.instr.read_until(message_terminator=bytes("\n", "utf-8"), timeout=0.001)

    def test_read_until_timeout(self):
        """Test read_until_timeout works exactly like read if data is available.

        Expecting:
            Timeout is set.
            Bytes read from VXI11 instrument has been passed untouched.
            Timeout is reverted.
        """
        self.mock().timeout = 10.0  # default timeout
        expected_read = b"a1b2c3d4"
        nbytes = len(expected_read)
        self.mock().read_raw.side_effect = [chr(c).encode() for c in expected_read]
        data = self.instr.read_until_timeout(nbytes, 1.5)
        self.assertAlmostEqual(self.mock().timeout, 10.0)
        self.assertEqual(nbytes, self.mock().read_raw.call_count)
        self.assertEqual(expected_read, data)

    def test_read_until_timeout_partial_read_with_timeout_exception(self):
        """Test read_until_timeout catches timeout exception and returns partial read from buffer.

        Expecting:
            Timeout is set.
            Read times out with QMI_TimeoutException.
            Bytes read from VXI11 instrument has been passed partially until the timeout.
            Second read can read the remaining bytes.
            Timeout is reverted.
        """
        self.mock().timeout = 10.0  # default timeout
        expected_read = b"a1b2c3d4"
        nbytes = len(expected_read)
        first_read_len = nbytes // 2
        second_read_len = nbytes - first_read_len
        self.mock().read_raw.side_effect = [chr(c).encode() for c in expected_read[:first_read_len]] +\
                                           [vxi11.vxi11.Vxi11Exception(err=15, note="Timeout!")] +\
                                           [chr(c).encode() for c in expected_read[first_read_len:]]
        # The first read should catch the timeout and return whatever was read until the timeout
        data_first_read = self.instr.read_until_timeout(nbytes, 1.5)
        data_second_read = self.instr.read_until_timeout(second_read_len, 1.5)
        data = data_first_read + data_second_read

        self.assertAlmostEqual(self.mock().timeout, 10.0)
        self.assertEqual(nbytes + 1, self.mock().read_raw.call_count)  # + 1 as exception also happens at read
        self.assertEqual(first_read_len, len(data_first_read))
        self.assertEqual(second_read_len, len(data_second_read))
        self.assertEqual(expected_read, data)

    def test_read_until_timeout_partial_read_with_endofinput_exception(self):
        """Test read_until_timeout excepts with QMI_EndOfInputException and read buffer remains as is.

        Expecting:
            Timeout is set.
            Read excepts with QMI_EndOfInputException.
            Bytes read from VXI11 instrument until exception remain in read buffer.
            Timeout is reverted.
        """
        self.mock().timeout = 10.0  # default timeout
        expected_read = b"a1b2c3d4"
        nbytes = len(expected_read)
        first_read_len = nbytes // 2
        self.mock().read_raw.side_effect = [chr(c).encode() for c in expected_read[:first_read_len]] +\
                                           [vxi11.vxi11.Vxi11Exception(note="End of input!")]
        # The read should except and return nothing.
        with self.assertRaises(qmi.core.exceptions.QMI_EndOfInputException):
            self.instr.read_until_timeout(nbytes, 1.5)

        self.assertAlmostEqual(self.mock().timeout, 10.0)
        self.assertEqual(first_read_len + 1, self.mock().read_raw.call_count)   # + 1 as exception also happens at read
        self.assertEqual(first_read_len, len(self.instr._read_buffer))

    def test_discard_read_vxi11exception(self):
        """Test discard_read command.

        Expecting:
            Data before timeout is discarded.
            Vxi11Exception number 15 only breaks the while loop.
            Data after timeout is not lost.
        """
        # Arrange
        discard_string = b"discard\r\n"
        test_string = b"test"
        self.mock().read_raw.side_effect = [bytearray(discard_string)] +\
            [vxi11.vxi11.Vxi11Exception(err=15, note="Test error!")] +\
            [test_string]
        # Act
        self.instr.discard_read()
        data = self.instr.read(len(test_string))
        # Assert
        self.assertEqual(test_string, data)

    def test_discard_read_timeouterror(self):
        """Test discard_read command.

        Expecting:
            Data before timeout is discarded.
            TimeoutError only breaks the while loop.
            Data after timeout is not lost.
        """
        # Arrange
        discard_string = b"discard\r\n"
        test_string = b"test"
        self.mock().read_raw.side_effect = [bytearray(discard_string)] +\
            [TimeoutError("Timeout!")] +\
            [test_string]
        # Act
        self.instr.discard_read()
        data = self.instr.read(len(test_string))
        # Assert
        self.assertEqual(test_string, data)

    def test_discard_read_exception(self):
        """Test discard_read raises a QMI_InstrumentException for Vxi11Exception != 15."""
        self.mock().read_raw.side_effect = vxi11.vxi11.Vxi11Exception(note="Test error!")
        with self.assertRaises(qmi.core.exceptions.QMI_InstrumentException):
            self.instr.discard_read()


class ListUsbtmcTransportsTestCase(unittest.TestCase):
    """Test the list_usbtmc_transports function."""
    @unittest.mock.patch('sys.platform', 'win32')
    def test_list_usbtmc_transports_win(self):
        # Expected resources found
        vendor_ids = [0x0699, 0xbebe]
        product_ids = [0x3000, 0xcafe]
        serial_nrs = ["XYZ", "ABC"]
        transport_str_1 = f"usbtmc:vendorid=0x{vendor_ids[0]:04x}:productid=0x{product_ids[0]:04x}:serialnr={serial_nrs[0]}"
        transport_str_2 = f"usbtmc:vendorid=0x{vendor_ids[1]:04x}:productid=0x{product_ids[1]:04x}:serialnr={serial_nrs[1]}"
        expected_resources = [transport_str_1, transport_str_2]
        # Act
        with unittest.mock.patch(
                "qmi.core.transport_usbtmc_visa.pyvisa.ResourceManager",
                return_value=ResourceManager()
        ):
            resources = list_usbtmc_transports()

        # Assert
        self.assertListEqual(expected_resources, resources)

    @unittest.mock.patch('sys.platform', 'linux1')
    def test_list_usbtmc_transports_lin(self):
        # Expected resources found
        vendor_ids = [0x0699, 0xbebe]
        product_ids = [0x3000, 0xcafe]
        serial_nrs = ["XYZ", "ABC"]
        transport_str_1 = f"usbtmc:vendorid=0x{vendor_ids[0]:04x}:productid=0x{product_ids[0]:04x}:serialnr={serial_nrs[0]}"
        transport_str_2 = f"usbtmc:vendorid=0x{vendor_ids[1]:04x}:productid=0x{product_ids[1]:04x}:serialnr={serial_nrs[1]}"
        expected_resources = [transport_str_1, transport_str_2]
        # Act
        with unittest.mock.patch(
                "qmi.core.transport_usbtmc_pyusb.usbtmc.list_resources",
                return_value=[visa_str_1, visa_str_3]
        ):
            resources = list_usbtmc_transports()

        # Assert
        self.assertListEqual(expected_resources, resources)


if __name__ == "__main__":
    unittest.main()



""" Testcase of the SDS1202X-E Siglent Oscilloscope. """
import unittest
from unittest.mock import call, patch
import numpy as np

import qmi
from qmi.instruments.siglent import Siglent_Sds1202xE, CommHeader, TriggerCondition
from qmi.core.transport import QMI_TcpTransport
from qmi.core.exceptions import QMI_InstrumentException, QMI_UsageException


class TestSDS1202XE(unittest.TestCase):
    """ Testcase of the TestSDS1202XE oscilloscope """

    def setUp(self):
        qmi.start("TestSiglentSDS1202X-E")
        # Add patches
        patcher = patch('qmi.instruments.siglent.sds1202xe.create_transport', spec=QMI_TcpTransport)
        self._transport_mock = patcher.start().return_value
        self.addCleanup(patcher.stop)
        patcher = patch('qmi.instruments.siglent.sds1202xe.ScpiProtocol', autospec=True)
        self._scpi_mock = patcher.start().return_value
        self.addCleanup(patcher.stop)
        # Make DUT
        self.scope: Siglent_Sds1202xE = qmi.make_instrument("SDS1202XE", Siglent_Sds1202xE, "")
        self.scope.open()

    def tearDown(self):
        self.scope.close()
        qmi.stop()

    def test_channel_exception(self):
        with self.assertRaises(ValueError):
            _ = self.scope.trace_dump(-1)
        with self.assertRaises(ValueError):
            _ = self.scope.trace_dump(6)

    def test_get_id(self):
        """ Test case for `get_id(...)` function. """
        # arrange
        expected_id_string = "test_string"
        expected_calls = [call.write('chdr off'), call.ask("*IDN?")]
        self._scpi_mock.ask.return_value = expected_id_string
        # act
        actual_val = self.scope.get_id()
        # assert
        self.assertEqual(self._scpi_mock.method_calls, expected_calls)
        self.assertEqual(actual_val, expected_id_string)

    def test_set_comm_header(self):
        """ Test case for `set_comm_header(...) method. """
        expected_scpi_calls = [
            [call.write("chdr long")],
            [call.write("chdr short")],
            [call.write("chdr off")],
        ]
        for i, header in enumerate(CommHeader):
            self._scpi_mock.reset_mock()
            with self.subTest(i=header.value):
                self.scope.set_comm_header(header)
                self.assertEqual(self._scpi_mock.method_calls, expected_scpi_calls[i])

    def test_get_voltage_per_division(self):
        """ Test case for `get_voltage_per_division(...)` method. """
        # arrange
        expected_vdiv = 1.0
        expected_calls = [call.write('chdr off'), call.ask("c1:vdiv?")]
        self._scpi_mock.ask.return_value = str(expected_vdiv)
        # act
        actual_val = self.scope.get_voltage_per_division(1)
        # assert
        self.assertEqual(self._scpi_mock.method_calls, expected_calls)
        self.assertEqual(actual_val, expected_vdiv)

    def test_get_voltage_per_division_exceptions(self):
        """ Test case for QMI_InstrumentException in `get_voltage_per_division(...)`"""
        # arrange
        self._scpi_mock.ask.side_effect = "Error"
        # act & assert
        with self.assertRaises(QMI_InstrumentException):
            _ = self.scope.get_voltage_per_division(1)

    def test_set_voltage_per_division(self):
        """ Test case for `set_voltage_per_division(...)` method. """
        # arrange
        expected_vdiv = 1.0
        expected_calls = [call.write('chdr off'), call.write(f"c1:vdiv {expected_vdiv}")]
        # act
        self.scope.set_voltage_per_division(1, expected_vdiv)
        # assert
        self.assertEqual(self._scpi_mock.method_calls, expected_calls)

    def test_set_voltage_per_division_exceptions(self):
        """ Test case for ValueError in `set_voltage_per_division(...)`"""
        # arrange
        out_of_range_vdivs = [100E-6, 20.0]
        # act & assert
        for vdiv in out_of_range_vdivs:
            with self.assertRaises(ValueError):
                self.scope.set_voltage_per_division(1, vdiv)

    def test_get_voltage_offset(self):
        """ Test case for `get_voltage_offset(...)` method. """
        # arrange
        expected_ofst = 1.0
        expected_calls = [call.write('chdr off'), call.ask("c1:ofst?")]
        self._scpi_mock.ask.return_value = str(expected_ofst)
        # act
        actual_val = self.scope.get_voltage_offset(1)
        # assert
        self.assertEqual(self._scpi_mock.method_calls, expected_calls)
        self.assertEqual(actual_val, expected_ofst)

    def test_get_voltage_offset_exceptions(self):
        """ Test case for QMI_InstrumentException in `get_voltage_offset(...)`"""
        # arrange
        self._scpi_mock.ask.side_effect = "Error"
        # act & assert
        with self.assertRaises(QMI_InstrumentException):
            self.scope.get_voltage_offset(1)

    def test_set_voltage_offset(self):
        """ Test case for `set_voltage_offset(...)` method. """
        # arrange
        expected_ofst = 1.0
        channel = 3
        expected_calls = [call.write('chdr off'), call.write(f"c3:ofst {expected_ofst}")]
        # act
        self.scope.set_voltage_offset(channel, expected_ofst)
        # assert
        self.assertEqual(self._scpi_mock.method_calls, expected_calls)

    def test_get_time_per_division(self):
        """ Test case for `get_time_per_division(...)` method. """
        # arrange
        expected_tdiv = 1.0
        expected_calls = [call.write('chdr off'), call.ask("tdiv?")]
        self._scpi_mock.ask.return_value = str(expected_tdiv)
        # act
        actual_val = self.scope.get_time_per_division()
        # assert
        self.assertEqual(self._scpi_mock.method_calls, expected_calls)
        self.assertEqual(actual_val, expected_tdiv)

    def test_get_time_per_division_exceptions(self):
        """ Test case for QMI_InstrumentException in `get_time_per_division(...)`"""
        # arrange
        self._scpi_mock.ask.side_effect = "Error"
        # act & assert
        with self.assertRaises(QMI_InstrumentException):
            self.scope.get_time_per_division()

    def test_set_time_per_division(self):
        """ Test case for `set_time_per_division(...)` method. """
        # arrange
        expected_tdivs = [1, 20, 50, 200]
        expected_units = ["s", "ms", "us", "ns"]
        expected_calls = [call.write('chdr off'),] + \
                         [call.write(f"tdiv {expected_tdivs[e]}{expected_units[e]}") for e in range(4)]
        # act
        for s in range(4):
            self.scope.set_time_per_division(expected_tdivs[s], expected_units[s])

        # assert
        self.assertEqual(self._scpi_mock.method_calls, expected_calls)

    def test_set_time_per_division_exceptions(self):
        """ Test case for ValueError in `set_time_per_division(...)`"""
        # arrange
        unexpected_tdiv = [1, 3, 200]
        unexpected_unit = ["ps", "ms", "s"]
        # act & assert
        for u in range(3):
            with self.assertRaises(ValueError):
                self.scope.set_time_per_division(unexpected_tdiv[u], unexpected_unit[u])

    def test_get_sample_rate(self):
        """ Test case for `get_sample_rate(...)` method. """
        # arrange
        expected_sara = 1.0
        expected_calls = [call.write('chdr off'), call.ask("sara?")]
        self._scpi_mock.ask.return_value = str(expected_sara)
        # act
        actual_val = self.scope.get_sample_rate()
        # assert
        self.assertEqual(self._scpi_mock.method_calls, expected_calls)
        self.assertEqual(actual_val, expected_sara)

    def test_get_sample_rate_exceptions(self):
        """ Test case for QMI_InstrumentException in `get_sample_rate(...)`"""
        # arrange
        self._scpi_mock.ask.side_effect = "Error"
        # act & assert
        with self.assertRaises(QMI_InstrumentException):
            _ = self.scope.get_sample_rate()

    def test_get_trigger_coupling(self):
        """ Test case for `get_trigger_coupling(...)` method. """
        # arrange
        expected_channel = 5
        expected_trcp = "2.0"
        expected_calls = [call.write('chdr off'), call.ask("ex5:trcp?")]
        self._scpi_mock.ask.return_value = expected_trcp
        # act
        actual_val = self.scope.get_trigger_coupling(expected_channel)
        # assert
        self.assertEqual(self._scpi_mock.method_calls, expected_calls)
        self.assertEqual(actual_val, expected_trcp)

    def test_set_trigger_coupling(self):
        """ Test case for `set_trigger_coupling(...)` method. """
        # arrange
        expected_trcps = ["ac", "dc", "hfrej", "lfrej"]
        channel = 5
        expected_calls = [call.write('chdr off'),] + [call.write(f"ex5:trcp {e}") for e in expected_trcps]
        # act
        for expected_trcp in expected_trcps:
            self.scope.set_trigger_coupling(channel, expected_trcp)

        # assert
        self.assertEqual(self._scpi_mock.method_calls, expected_calls)

    def test_set_trigger_coupling_exception(self):
        """ Test case for ValueError in `set_trigger_coupling(...) method."""
        # arrange
        unexpected_trcp = "ac/dc"
        # act
        with self.assertRaises(ValueError) as exc:
            self.scope.set_trigger_coupling(3, unexpected_trcp)

        self.assertIn(unexpected_trcp, str(exc.exception))

    def test_get_trigger_level(self):
        """ Test case for `get_trigger_level(...)` method. """
        # arrange
        expected_channel = 4
        expected_trlv = 5.5
        expected_calls = [call.write('chdr off'), call.ask("c4:trlv?")]
        self._scpi_mock.ask.return_value = str(expected_trlv)
        # act
        actual_val = self.scope.get_trigger_level(expected_channel)
        # assert
        self.assertEqual(self._scpi_mock.method_calls, expected_calls)
        self.assertEqual(actual_val, expected_trlv)

    def test_set_trigger_level_external(self):
        """ Test case for `set_trigger_level(...)` method for an external channel. """
        # arrange
        expected_trlv = 2.0
        channel = 5
        expected_calls = [call.write('chdr off'), call.write(f"ex5:trlv {expected_trlv}")]
        # act
        self.scope.set_trigger_level(channel, expected_trlv)

        # assert
        self.assertEqual(self._scpi_mock.method_calls, expected_calls)

    def test_set_trigger_level_internal(self):
        """ Test case for `set_trigger_level(...)` method for an internal channel. """
        # arrange
        expected_vdiv = "1.0"
        expected_trlv = 2.0
        channel = 4
        expected_calls = [call.write('chdr off'), call.write(f"c4:trlv {expected_trlv}")]
        self._scpi_mock.ask.return_value = expected_vdiv
        # act
        self.scope.set_trigger_level(channel, expected_trlv)

        # assert
        self.assertEqual(self._scpi_mock.method_calls, expected_calls)

    def test_get_trigger_mode(self):
        """ Test case for `get_trigger_mode(...)` method. """
        # arrange
        expected_trmd = "AUTO"
        expected_calls = [call.write('chdr off'), call.ask("trmd?")]
        self._scpi_mock.ask.return_value = expected_trmd
        # act
        actual_val = self.scope.get_trigger_mode()
        # assert
        self.assertEqual(self._scpi_mock.method_calls, expected_calls)
        self.assertEqual(actual_val, expected_trmd)

    def test_set_trigger_mode(self):
        """ Test case for `set_trigger_mode(...)` method. """
        # arrange
        expected_trmds = ["auto", "norm", "single", "STOP"]
        expected_calls = [call.write('chdr off'),] + [call.write(f"trmd {e}") for e in expected_trmds]
        # act
        for expected_trmd in expected_trmds:
            self.scope.set_trigger_mode(expected_trmd)

        # assert
        self.assertEqual(self._scpi_mock.method_calls, expected_calls)

    def test_set_trigger_mode_exception(self):
        """ Test case for ValueError in `set_trigger_mode(...) method."""
        # arrange
        unexpected_trmd = "DOUBLE"
        # act
        with self.assertRaises(ValueError) as exc:
            self.scope.set_trigger_mode(unexpected_trmd)

        self.assertIn(unexpected_trmd, str(exc.exception))

    def test_get_trigger_select(self):
        """ Test case for `get_trigger_select(...)` method. """
        # arrange
        trmd1 = "trig_type,source,hold_type,hold_value,hold_value2"
        expected_trmd1 = TriggerCondition("trig_type", "source", "hold_type", "hold_value", "hold_value2")
        expected_calls = [call.write('chdr off'), call.ask("trse?")]
        self._scpi_mock.ask.return_value = trmd1
        # act
        actual_val1 = self.scope.get_trigger_select()
        # assert
        self.assertEqual(self._scpi_mock.method_calls, expected_calls)
        self.assertEqual(actual_val1, expected_trmd1)

    def test_get_trigger_slope(self):
        """ Test case for `get_trigger_slope(...)` method. """
        # arrange
        expected_trsl = "window"
        expected_calls = [call.write('chdr off'), call.ask("c1:trsl?")]
        self._scpi_mock.ask.return_value = expected_trsl
        # act
        actual_val = self.scope.get_trigger_slope(1)
        # assert
        self.assertEqual(self._scpi_mock.method_calls, expected_calls)
        self.assertEqual(actual_val, expected_trsl)

    def test_set_trigger_slope(self):
        """ Test case for `set_trigger_slope(...)` method. """
        # arrange
        expected_trsls = ["neg", "POS", "window"]
        channels = [3, 2, 1]
        expected_calls = [call.write('chdr off'),] + \
                         [call.write(f"c{c}:trsl {e}") for c, e in zip(channels, expected_trsls)]
        # act
        for channel, expected_trsl in zip(channels, expected_trsls):
            self.scope.set_trigger_slope(channel, expected_trsl)

        # assert
        self.assertEqual(self._scpi_mock.method_calls, expected_calls)

    def test_set_trigger_slope_exception(self):
        """ Test case for ValueError in `set_trigger_slope(...) method."""
        # arrange
        unexpected_trsl = "neutral"
        # act
        with self.assertRaises(ValueError) as exc:
            self.scope.set_trigger_slope(0, unexpected_trsl)

        self.assertIn(unexpected_trsl, str(exc.exception))

    def test_arm_trigger(self):
        """Test for `arm_trigger()` method."""
        # arrange
        expected_calls = [call.write('chdr off'), call.write("arm")]
        # act
        self.scope.arm_trigger()
        # assert
        self.assertEqual(self._scpi_mock.method_calls, expected_calls)

    def test_stop_acquisition(self):
        """Test for `stop_acquisition()` method."""
        # arrange
        expected_calls = [call.write('chdr off'), call.write("stop")]
        # act
        self.scope.stop_acquisition()
        # assert
        self.assertEqual(self._scpi_mock.method_calls, expected_calls)

    def test_get_trigger_state(self):
        """Test how to get trigger state with `get_trigger_state()` method."""
        # arrange
        expected_state = 0
        expected_calls = [call.write('chdr off'), call.ask("inr?")]
        self._scpi_mock.ask.return_value = str(expected_state)
        # act
        state = self.scope.get_trigger_state()
        # assert
        self.assertEqual(self._scpi_mock.method_calls, expected_calls)
        self.assertEqual(state, expected_state)

    def test_get_trigger_state_exception(self):
        """Test getting trigger state with `get_trigger_state()` method throws exception."""
        # arrange
        expected_state = "zero"
        self._scpi_mock.ask.return_value = expected_state
        # act
        with self.assertRaises(QMI_InstrumentException):
            self.scope.get_trigger_state()

    def test_screen_dump(self):
        """ Test case for `screen_dump(...)` function. """
        # arrange
        expected_dump = bytes(768066)  # replace with random string
        expected_calls_scpi = [call.write('chdr off'), call.write("SCDP")]
        expected_calls_transport = [
            call.open,
            call.read(len(expected_dump), 2.0),  # can you make an argument a don't care?
            call.read(1, 2.0)
        ]
        self._transport_mock.read.side_effect = [expected_dump, b'\n']
        # act
        actual_dump = self.scope.screen_dump()
        # assert
        self.assertEqual(self._scpi_mock.method_calls, expected_calls_scpi)
        self.assertEqual(self._transport_mock.method_calls, expected_calls_transport)
        self.assertEqual(actual_dump, expected_dump)

    def test_screen_dump_exception(self):
        """ Test case for `screen_dump(...)` exception handling. """
        # arrange
        self._transport_mock.read.side_effect = [bytes(768066), b'\x00']
        # act
        with self.assertRaises(QMI_InstrumentException):
            _ = self.scope.screen_dump()

    def test_get_waveform_channel_1(self):
        """ Test case for `get_waveform(...)` function. """
        # arrange
        expected_dump = b'Hello World!'  # replace with random string
        expected_calls_scpi = [
            call.write('chdr off'),
            call.write("c1:wf? dat2"),
            call.read_binary_data()
        ]
        expected_calls_transport = [
            call.open,
            call.read(5, 2.0),
            call.read(1, 2.0)
        ]
        self._transport_mock.read.side_effect = [b"DAT2,", b'\n']
        self._scpi_mock.read_binary_data.return_value = expected_dump
        # act
        actual_dump = self.scope.get_waveform(1)
        # assert
        self.assertEqual(self._scpi_mock.method_calls, expected_calls_scpi)
        self.assertEqual(self._transport_mock.method_calls, expected_calls_transport)
        self.assertEqual(actual_dump, expected_dump)

    def test_get_waveform_channel_digital(self):
        """ Test case for `get_waveform(...)` function. """
        # arrange
        expected_dump = b'Hello World!'  # replace with random string
        expected_calls_scpi = [
            call.write('chdr off'),
            call.write("d1:wf? dat2"),
            call.read_binary_data()
        ]
        expected_calls_transport = [
            call.open,
            call.read(5, 2.0),
            call.read(1, 2.0)
        ]
        self._transport_mock.read.side_effect = [b"DAT2,", b'\n']
        self._scpi_mock.read_binary_data.return_value = expected_dump
        # act
        actual_dump = self.scope.get_waveform("d1")
        # assert
        self.assertEqual(self._scpi_mock.method_calls, expected_calls_scpi)
        self.assertEqual(self._transport_mock.method_calls, expected_calls_transport)
        self.assertEqual(actual_dump, expected_dump)

    def test_get_waveform_exceptions_invalid_response(self):
        """ Test case for `get_waveform(...)` exception handling. """
        # arrange
        side_effects = {
            "invalid_hdr": [b'\x00'],
            "invalid_trm": [b"DAT2,", b'\x00']
        }
        # act
        for k, v in side_effects.items():
            with self.subTest(i=k):
                self._transport_mock.read.side_effect = v
                with self.assertRaises(QMI_InstrumentException):
                    self.scope.get_waveform(1)

    def test_get_waveform_exceptions_invalid_channel(self):
        """ Test case for `get_waveform(...)` exception handling. """
        # arrange
        side_effects = {
            "invalid_hdr": [b'\x00'],
            "invalid_trm": [b"DAT2,", b'\x00']
        }
        # act
        for k, v in side_effects.items():
            with self.subTest(i=k):
                self._transport_mock.read.side_effect = v
                with self.assertRaises(QMI_UsageException):
                    self.scope.get_waveform("d100")

    def test_get_waveform_exceptions_invalid_channel_type(self):
        """ Test case for `get_waveform(...)` exception handling. """
        # arrange
        side_effects = {
            "invalid_hdr": [b'\x00'],
            "invalid_trm": [b"DAT2,", b'\x00']
        }
        # act
        for k, v in side_effects.items():
            with self.subTest(i=k):
                self._transport_mock.read.side_effect = v
                with self.assertRaises(QMI_UsageException):
                    self.scope.get_waveform(1.1)

    def test_trace_dump(self):
        """ Test case for `trace_dump(...)` function. This test mainly checks the logic of the conversion from 
            byte data to actual time and voltage data which are returned as np.ndarrays. 
            Logic from datasheet are as follows:
                - voltage value (V) = code value *(vdiv /25) - voffset
                - time value(S) = - index*(tdiv*grid/2).
        """
        # arrange
        channel = 1
        expected_trace = (
            np.arange(-7.0, 8.0, 1.0),
            np.full((15,), 2.0)
        )
        expected_calls_scpi = [
            call.write('chdr off'),
            call.write("chdr off"),
            call.ask(f"c{channel}:vdiv?"),
            call.ask(f"c{channel}:ofst?"),
            call.ask("tdiv?"),
            call.ask("sara?"),
            call.write(f"c{channel}:wf? dat2"),
            call.read_binary_data(),
        ]
        self._scpi_mock.ask.side_effect = ['1.0', '-1.0', '1.0', '1.0']
        self._scpi_mock.read_binary_data.return_value = bytes([25 for _ in range(15)])
        expected_calls_transport = [
            call.open,
            call.discard_read(),
            call.read(5, 2.0),
            call.read(1, 2.0),
        ]
        self._transport_mock.read.side_effect = [b'DAT2,', b'\n']
        # act
        actual_dump = self.scope.trace_dump(channel)
        # assert
        self.assertEqual(self._scpi_mock.method_calls, expected_calls_scpi)
        self.assertEqual(self._transport_mock.method_calls, expected_calls_transport)
        self.assertEqual(actual_dump[0].tolist(), expected_trace[0].tolist())
        self.assertEqual(actual_dump[1].tolist(), expected_trace[1].tolist())


if __name__ == '__main__':
    unittest.main()



"""Unit tests for a Newport single axis motion controller."""
from typing import cast
import unittest
from unittest.mock import Mock, call, patch

import qmi
from qmi.core.exceptions import QMI_InstrumentException
from qmi.core.transport import QMI_TcpTransport
from qmi.instruments.newport import Newport_Smc100Pp
from qmi.instruments.newport.actuators import UTS100PP


class TestNewportSmc100Pp(unittest.TestCase):
    """
    Tests for the single axis motion controller.
    """

    TRANSPORT_STR = "/dev/cu.usbserial-FT5TMFGL"

    def setUp(self):
        qmi.start("Test100PpControllerContext", console_loglevel="CRITICAL")
        # Add patches
        patcher = patch(
            'qmi.instruments.newport.single_axis_motion_controller.create_transport', spec=QMI_TcpTransport)
        self._transport_mock: Mock = patcher.start().return_value
        self.addCleanup(patcher.stop)
        patcher = patch(
            'qmi.instruments.newport.single_axis_motion_controller.ScpiProtocol', autospec=True)
        self._scpi_mock: Mock = patcher.start().return_value
        self.addCleanup(patcher.stop)
        # Make DUTs
        self.controller_address = 1
        Newport_Smc100Pp.PW0_EXEC_TIME = 0.1
        Newport_Smc100Pp.COMMAND_EXEC_TIME = 0.01
        self.instr: Newport_Smc100Pp = qmi.make_instrument(
            "sam_controller", Newport_Smc100Pp, self.TRANSPORT_STR, "FT5TMFGL",
            {1: UTS100PP},
            90210)
        self.instr = cast(Newport_Smc100Pp, self.instr)
        self.instr.open()

    def tearDown(self):
        self.instr.close()
        qmi.stop()

    def test_get_micro_step_per_full_step_factor(self):
        """Test getting motor micro-step factor."""
        state = "14"  # Only need to exit from CONFIGURATION mode.
        expected_value = 24
        self._scpi_mock.ask.side_effect = [f"1TS0000{state}", f"FRM {expected_value}", "@"]
        expected_write_calls = [
            call(f"{self.controller_address}PW0\r\n")
        ]
        expected_ask_calls = [
            call(f"{self.controller_address}TS\r\n"),
            call(f"{self.controller_address}FRM?\r\n"),
            call(f"{self.controller_address}TE\r\n")
        ]

        actual_value = self.instr.get_micro_step_per_full_step_factor()

        self.assertEqual(actual_value, expected_value)

        self._scpi_mock.write.assert_has_calls(expected_write_calls)
        self._scpi_mock.ask.assert_has_calls(expected_ask_calls)

    def test_set_micro_step_per_full_step_factor(self):
        """Test setting motor micro-step factor."""
        state = "14"  # Only need to exit from CONFIGURATION mode.
        micro_step_per_full_step_factor = 24
        self._scpi_mock.ask.side_effect = [f"1TS0000{state}", "@"]
        expected_write_calls = [
            call(f"1FRM{micro_step_per_full_step_factor}\r\n"),
            call(f"{self.controller_address}PW0\r\n")
        ]
        expected_ask_calls = [
            call(f"{self.controller_address}TS\r\n"),
            call(f"{self.controller_address}TE\r\n")
        ]

        self.instr.set_micro_step_per_full_step_factor(micro_step_per_full_step_factor)

        self._scpi_mock.write.assert_has_calls(expected_write_calls)
        self._scpi_mock.ask.assert_has_calls(expected_ask_calls)

    def test_set_micro_step_per_full_step_factor_excepts(self):
        """Test setting motor micro-step factor excepts when value out of valid range."""
        micro_step_per_full_step_factor_invalid = [0, 2001]
        for value in micro_step_per_full_step_factor_invalid:
            with self.assertRaises(QMI_InstrumentException):
                self.instr.set_micro_step_per_full_step_factor(value)

        self._scpi_mock.write.assert_not_called()
        self._scpi_mock.ask.assert_not_called()

    def test_get_motion_distance_per_full_step(self):
        """Test getting motor motion distance."""
        state = "14"  # Only need to exit from CONFIGURATION mode.
        expected_value = 24
        self._scpi_mock.ask.side_effect = [f"1TS0000{state}", f"FRS {expected_value}", "@"]
        expected_write_calls = [
            call(f"{self.controller_address}PW0\r\n")
        ]
        expected_ask_calls = [
            call(f"{self.controller_address}TS\r\n"),
            call(f"{self.controller_address}FRS?\r\n"),
            call(f"{self.controller_address}TE\r\n")
        ]

        actual_value = self.instr.get_motion_distance_per_full_step()
        self.assertEqual(actual_value, expected_value)

        self._scpi_mock.write.assert_has_calls(expected_write_calls)
        self._scpi_mock.ask.assert_has_calls(expected_ask_calls)

    def test_set_motion_distance_per_full_step(self):
        """Test setting motor motion distance."""
        state = "14"  # Only need to exit from CONFIGURATION mode.
        motion_distance_per_full_step = 24
        self._scpi_mock.ask.side_effect = [f"1TS0000{state}", "@"]
        expected_write_calls = [
            call(f"1FRS{motion_distance_per_full_step}\r\n"),
            call(f"{self.controller_address}PW0\r\n")
        ]
        expected_ask_calls = [
            call(f"{self.controller_address}TS\r\n"),
            call(f"{self.controller_address}TE\r\n")
        ]

        self.instr.set_motion_distance_per_full_step(motion_distance_per_full_step)

        self._scpi_mock.write.assert_has_calls(expected_write_calls)
        self._scpi_mock.ask.assert_has_calls(expected_ask_calls)

    def test_set_motion_distance_per_full_step_excepts(self):
        """Test setting motor motion distance excepts when value out of valid range."""
        motion_distance_per_full_step_invalid = [1E-6, 1E12]
        for value in motion_distance_per_full_step_invalid:
            with self.assertRaises(QMI_InstrumentException):
                self.instr.set_motion_distance_per_full_step(value)

        self._scpi_mock.write.assert_not_called()
        self._scpi_mock.ask.assert_not_called()

    def test_get_base_velocity(self):
        """Test getting base velocity."""
        state = "3C"  # Set to DISABLE state to take the shortest path
        expected_vel = 1.5
        self._scpi_mock.ask.side_effect = [
            f"1TS0000{state}", f"{self.controller_address}VB %s" % expected_vel, "@"]
        expected_ask_calls = [
            call(f"{self.controller_address}TS\r\n"),
            call(f"{self.controller_address}VB?\r\n"),
            call(f"{self.controller_address}TE\r\n")
        ]
        expected_write_calls = [
            call(f"{self.controller_address}MM0\r\n")
        ]

        actual_value = self.instr.get_base_velocity()

        self.assertEqual(actual_value, expected_vel)
        self._scpi_mock.write.assert_has_calls(expected_write_calls)
        self._scpi_mock.ask.assert_has_calls(expected_ask_calls)

    def test_get_base_velocity_from_configuration(self):
        """Test getting base velocity from CONFIGURATION state."""
        state = "14"  # Set to CONFIGURATION state
        expected_vel = 1.5
        self._scpi_mock.ask.side_effect = [
            f"1TS0000{state}", f"1TS0000{state}", f"{self.controller_address}VB %s" % expected_vel, "@"]
        expected_ask_calls = [
            call(f"{self.controller_address}TS\r\n"),
            call(f"{self.controller_address}TS\r\n"),
            call(f"{self.controller_address}VB?\r\n"),
            call(f"{self.controller_address}TE\r\n")
        ]
        expected_write_calls = [
            call(f"{self.controller_address}PW0\r\n")
        ]

        actual_value = self.instr.get_base_velocity()

        self.assertEqual(actual_value, expected_vel)
        self._scpi_mock.write.assert_has_calls(expected_write_calls)
        self._scpi_mock.ask.assert_has_calls(expected_ask_calls)

    def test_set_base_velocity_with_persist(self):
        """Test setting base velocity distance with persist."""
        state = "14"  # Set to CONFIGURATION state to take the shortest path
        persist = True
        base_velocity = 24
        max_vel = 1000
        self._scpi_mock.ask.side_effect = [f"1TS0000{state}", f"1VA{max_vel}", "@", "@"]
        expected_write_calls = [
            call(f"{self.controller_address}VB%s\r\n" % base_velocity),
            call(f"{self.controller_address}PW0\r\n")
        ]
        expected_ask_calls = [
            call(f"{self.controller_address}TS\r\n"),
            call(f"{self.controller_address}VA?\r\n"),
            call(f"{self.controller_address}TE\r\n"),
        ]

        self.instr.set_base_velocity(base_velocity, persist)

        self._scpi_mock.write.assert_has_calls(expected_write_calls)
        self._scpi_mock.ask.assert_has_calls(expected_ask_calls)

    def test_set_base_velocity(self):
        """Test setting base velocity distance."""
        state = "34"  # Set to READY state to take the shortest path
        base_velocity = 24
        max_vel = 1000
        self._scpi_mock.ask.side_effect = [f"1TS0000{state}", f"1VA{max_vel}", "@", "@"]
        expected_write_calls = [
            call(f"{self.controller_address}VB%s\r\n" % base_velocity)
        ]
        expected_ask_calls = [
            call(f"{self.controller_address}TS\r\n"),
            call(f"{self.controller_address}VA?\r\n"),
            call(f"{self.controller_address}TE\r\n"),
        ]

        self.instr.set_base_velocity(base_velocity)

        self._scpi_mock.write.assert_has_calls(expected_write_calls)
        self._scpi_mock.ask.assert_has_calls(expected_ask_calls)

    def test_set_base_velocity_excepts_with_persist(self):
        """Test setting motor motion distance excepts when value out of valid range."""
        state = "14"  # Set to CONFIGURATION state to take the shortest path
        persist = True
        max_vel = 1000
        self._scpi_mock.ask.side_effect = [
            f"1TS0000{state}", f"1VA{max_vel}", "@", f"1TS0000{state}", f"1VA{max_vel}", "@"
        ]
        expected_write_calls = [
            call(f"{self.controller_address}PW0\r\n"),
            call(f"{self.controller_address}PW0\r\n")
        ]
        expected_ask_calls = [
            call(f"{self.controller_address}TS\r\n"),
            call(f"{self.controller_address}VA?\r\n"),
            call(f"{self.controller_address}TE\r\n"),
            call(f"{self.controller_address}TS\r\n"),
            call(f"{self.controller_address}VA?\r\n"),
            call(f"{self.controller_address}TE\r\n")
        ]

        base_velocity_invalid = [-0.001, max_vel + 0.1]
        for value in base_velocity_invalid:
            with self.assertRaises(QMI_InstrumentException):
                self.instr.set_base_velocity(value, persist)

        self._scpi_mock.write.assert_has_calls(expected_write_calls)
        self._scpi_mock.ask.assert_has_calls(expected_ask_calls)

    def test_set_base_velocity_excepts(self):
        """Test setting motor motion distance excepts when value out of valid range."""
        state = "34"  # Set to READY state to take the shortest path
        max_vel = 1000
        self._scpi_mock.ask.side_effect = [
            f"1TS0000{state}", f"1VA{max_vel}", "@", f"1TS0000{state}", f"1VA{max_vel}", "@"
        ]
        expected_ask_calls = [
            call(f"{self.controller_address}TS\r\n"),
            call(f"{self.controller_address}VA?\r\n"),
            call(f"{self.controller_address}TE\r\n"),
            call(f"{self.controller_address}TS\r\n"),
            call(f"{self.controller_address}VA?\r\n"),
            call(f"{self.controller_address}TE\r\n")
        ]

        base_velocity_invalid = [-0.001, max_vel + 0.1]
        for value in base_velocity_invalid:
            with self.assertRaises(QMI_InstrumentException):
                self.instr.set_base_velocity(value)

        self._scpi_mock.write.assert_not_called()
        self._scpi_mock.ask.assert_has_calls(expected_ask_calls)



"""
Jäger Computergesteuerte Messtechnik GmbH, 'ADwin' devices.

The qmi.instruments.adwin package provides support for:
- high-capacity compact real-time systems
  - ADwin-Gold II
  - ADwin-Pro II
- T11, T12 and T12.1 processor types.
"""
# Alternative, QMI naming convention approved names
from qmi.instruments.adwin.proii import Adwin_ProII as JagerMessTechnik_AdwinProII
from qmi.instruments.adwin.goldii import Adwin_GoldII as JagerMessTechnik_AdwinGoldII
